import random
import math
import numpy as np
from typing import List, Tuple, Optional, FrozenSet, Union
import open3d as o3d

def __set_random_seed(seed):
    """
    Set the random seed for reproducibility.

    :param seed: int
        The seed value for the random number generator.
    """
    random.seed(seed)

def __sample_point_circle_2d(center: Tuple[float, float], radius: float=1) -> Tuple[float, float]:
    """
    Generate a random point on a 2D circle with a specified radius and center, using rejection sampling.

    :param center: tuple
        A tuple (x, y) representing the center of the circle.
    :param radius: float, optional
        The radius of the circle. Default is 1.
    :return: tuple
        A tuple (x, y) representing the coordinates of the sampled point.
    """
    if radius <= 0:
        raise ValueError("Radius must be positive.")
    while True:
        # Generate a random point on the circle with radius 1 centered at (0,0)
        x = random.uniform(-1, 1)
        y = random.uniform(-1, 1)
        
        # Scale the coordinates to the desired radius
        x *= radius
        y *= radius
        
        # Translate the point to the desired center
        x += center[0]
        y += center[1]
        
        # Check if the point is within the circle
        if (x - center[0])**2 + (y - center[1])**2 <= radius**2:
            return x, y

def sampling_circle_2d(n_samples:int=1, center:Tuple[float, float]=(0,0), radius:float=1, seed:Optional[int]=None):
    """
    Generate random samples on a 2D circle with a specified radius and center.

    :param n_samples: The number of random samples to generate on the circle.
    :type n_samples: int
    :param center: A tuple (x, y) representing the center of the circle. Default is (0,0).
    :type center: Tuple[float, float]
    :param radius: The radius of the circle. Default is 1.
    :type radius: float
    :param seed: The seed value for the random number generator. Default is None.
    :type seed: int
    :return: A list of tuples (x, y) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float]]
        
    .. note:: The samples are uniformly distributed on the circle.
    
    Examples
    --------
    Generate 100 random samples on a circle with center (2,3) and radius 5:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geometry
    >>> import matplotlib.pyplot as plt

    Define the limits of the parallelogram and the number of points to sample:

    >>> n_samples = 100
    >>> center = (2,3)
    >>> radius = 5

    Generate the samples:
    
    >>> samples = sampling.sampling_circle_2d(n_samples=100, center=center, radius=radius, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [(3.3942679845788373, -1.7498924477733304),
    (-0.24970681630880742, 0.23210738148822774),
    (4.364712141640124, 4.766994874229113),
    (1.2192181968527043, -1.7020278056192968),
    (-0.8136202519639664, 3.0535528810336237)]

    Plot the samples:
    
    >>> fig, ax = plt.subplots()
    >>> xlim_min = center[0] - radius
    >>> xlim_max = center[0] + radius
    >>> ylim_min = center[1] - radius
    >>> ylim_max = center[1] + radius
    >>> graph_limits = geometry.get_limits_of_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max)
    >>> ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    >>> ax.set_ylim(graph_limits[2], graph_limits[3]*1.1)  # Set y-axis limits
    >>> ax.scatter(*zip(*samples))
    >>> ax.set_aspect('equal')
    >>> # create title from n_samples, center, and radius, using f-string
    >>> title = (f'{n_samples} Samples on a Circle with Center {center} and Radius {radius}')
    >>> ax.set_title(title)
    >>> # draw also the circle in red
    >>> circle = plt.Circle(center, radius, color='r', fill=False)
    >>> ax.add_artist(circle)
    >>> # draw the X and Y axes in dotted lines
    >>> ax.axhline(0, linestyle='dotted', color='black')
    >>> ax.axvline(0, linestyle='dotted', color='black')
    >>> ax.set_xlabel('X')
    >>> ax.set_ylabel('Y')
    >>> plt.show()

    |sampling_circle_2d|

    .. |sampling_circle_2d| image:: ../../_static/images/sampling_circle_2d.png
    """
    # Set the random seed for reproducibility if provided
    if seed is not None:
        __set_random_seed(seed)

    # Generate the specified number of samples on the circle
    samples = [__sample_point_circle_2d(center, radius) for _ in range(n_samples)]

    return samples
        
def __sample_point_circle_3d(radius: float = 1.0, center: np.ndarray =np.array([0, 0, 0]), normal: np.ndarray=np.array([0, 0, 1])) -> np.ndarray:
    '''
    Sample a point from a 3D circle using the rejection sampling method.

    :param radius: The radius of the circle.
    :type radius: float
    :param center: The center of the circle.
    :type center: np.ndarray
    :param normal: The normal vector of the plane on which the circle lies.
    :type normal: np.ndarray
    :return: A list of tuples (x, y, z) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float, float]]

    .. note:: The samples are uniformly distributed on the circle.
    
    '''
    normal = normal / np.linalg.norm(normal)
    
    while True:
        # Generate a random point within the bounding box
        point = center + np.array([random.uniform(-radius, radius) for _ in range(3)])

        # Project the point onto the plane defined by the circle
        projected_point = point - np.dot(point - center, normal) * normal

        # Check if the projected point lies within the circle
        if np.linalg.norm(projected_point - center) <= radius:
            return projected_point


def sampling_circle_3d(n_samples, radius: float = 1.0, center: np.ndarray =np.array([0, 0, 0]), 
                        normal: np.ndarray=np.array([0, 0, 1]), seed: Optional[int]=None) -> List[np.ndarray]:
    '''
    Sample a n_samples number of points from a 3D circle using the rejection sampling method.

    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param radius: The radius of the circle.
    :type radius: float
    :param center: The center of the circle.
    :type center: np.ndarray
    :param normal: The normal vector of the plane on which the circle lies.
    :type normal: np.ndarray
    :return: A list of tuples (x, y, z) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float, float]]

    .. note:: The samples are uniformly distributed on the circle.

    :Example:

    ::

    Sample 100 points from a circle in an arbitrary position and with arbitrary normal vector:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geom
    >>> import mrdja.matplot3d as plt3d
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np

    Define the parameters of the circle and the number of points to sample:

    >>> n_samples = 100
    >>> radius = 5
    >>> center = np.array([1, 2, 0])
    >>> normal = np.array([1, 1, 0]) / np.linalg.norm(np.array([1, 1, 0]))

    Sample the points:

    >>> samples = sampling.sampling_circle_3d(n_samples=n_samples, radius=radius, center=center, normal=normal, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [array([ 4.07208022, -1.07208022, -2.24970682]),
    array([-1.56630238,  4.56630238,  1.76699487]),
    array([0.05579622, 2.94420378, 0.05355288]),
    array([0.13849159, 2.86150841, 1.49884438]),
    array([2.62250429, 0.37749571, 0.89265684])]

    Plot the 3D circle and the samples:
    
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection='3d')
    >>> plt3d.draw_circumference(center=center, radius=radius, normal=normal, color="blue", alpha=0.2, ax=ax)
    >>> xlim_min = np.min([sample[0] for sample in samples])
    >>> xlim_max = np.max([sample[0] for sample in samples])
    >>> ylim_min = np.min([sample[1] for sample in samples])
    >>> ylim_max = np.max([sample[1] for sample in samples])
    >>> zlim_min = np.min([sample[2] for sample in samples])
    >>> zlim_max = np.max([sample[2] for sample in samples])
    >>> graph_limits = geom.get_limits_of_3d_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max, zlim_min, zlim_max)
    >>> ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    >>> ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
    >>> ax.set_zlim(graph_limits[4], graph_limits[5])  # Set z-axis limits
    >>> ax.scatter(*zip(*samples))
    >>> # create title from n_samples, center, and radius, using f-string
    >>> title = (f'{n_samples} Samples on a Circle with Center {center} and Radius {radius}')
    >>> ax.set_title(title)
    >>> ax.set_xlabel('X')
    >>> ax.set_ylabel('Y')
    >>> ax.set_zlabel('Z')
    >>> plt.show()

    |sampling_circle_3d|

    .. |sampling_circle_3d| image:: ../../_static/images/sampling_circle_3d.png

    '''
    if seed is not None:
        random.seed(seed)
    samples = [__sample_point_circle_3d(radius, center, normal) for _ in range(n_samples)]
    return samples

def __sample_point_parallelogram_2d(normal1: Tuple[float, float], normal2: Tuple[float, float], center: Tuple[float, float], length1: float, length2: float) -> Tuple[float, float]:
    '''
    Sample a point from a parallelogram with sides parallel to the vectors normal1 and normal2.

    :param normal1: Tuple[float, float]
        The first vector normal to the sides of the parallelogram.
    :param normal2: Tuple[float, float]
        The second vector normal to the sides of the parallelogram.
    :param center: Tuple[float, float]
        The center of the parallelogram.
    :param length1: float
        The length of the first side of the parallelogram.
    :param length2: float
        The length of the second side of the parallelogram.
    :return: Tuple[float, float]
        A tuple (x, y) representing the coordinates of the sampled point.
    '''
    # Generate two random numbers between -length1/2 and length1/2 and -length2/2 and length2/2 respectively
    x = random.uniform(-length1/2, length1/2)
    y = random.uniform(-length2/2, length2/2)
    # Those numbers represent the coordinates of the sampled point in the normal coordinate system
    # Transform the coordinates to the global coordinate system
    projected_point = (x * normal1[0] + y * normal2[0] + center[0], x * normal1[1] + y * normal2[1] + center[1])
    return projected_point

def sampling_parallelogram_2d(n_samples: int, normal1: Tuple[float, float], normal2: Tuple[float, float], 
                              center: Tuple[float, float], length1: float, length2: float, 
                              seed: Optional[int] = None) -> List[Tuple[float, float]]:
    """
    Sample a n_samples number of points from a 2D parallelogram.
     
    The parallelogram has sides parallel to the vectors normal1 and normal2, 
    with lengths length1 and length2 respectively, and centered at center.

    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param normal1: The first vector normal to the sides of the parallelogram.
    :type normal1: Tuple[float, float]
    :param normal2: The second vector normal to the sides of the parallelogram.
    :type normal2: Tuple[float, float]
    :param center: The center of the parallelogram.
    :type center: Tuple[float, float]
    :param length1: The length of the first side of the parallelogram.
    :type length1: float
    :param length2: The length of the second side of the parallelogram.
    :type length2: float
    :return: A list of tuples (x, y) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float]]
        
    .. note:: The samples are uniformly distributed on the parallelogram.

    Examples
    --------
    Sample 100 points from a parallelogram in an arbitrary position and with arbitrary sides:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geometry
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np

    Define the parameters of the parallelogram and the number of points to sample:

    >>> n_samples = 100
    >>> normal1 = (1, 1)
    >>> normal2 = (-2, 1)
    >>> center = (1, 2)
    >>> length1 = 5
    >>> length2 = 4

    Sample the points:

    >>> samples = sampling.sampling_parallelogram_2d(n_samples=n_samples, normal1=normal1, normal2=normal2, 
                                                center=center, length1=length1, length2=length2, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [(5.497047950508083, 0.7971770131800864),
    (2.0894606866550145, -0.23201045555911293),
    (0.7687601714367718, 3.8891540205117074),
    (6.265387177488898, 2.3086531690418917),
    (4.3712313429217895, -0.2712020238213664)]

    Plot the samples:

    >>> fig, ax = plt.subplots()
    >>> ax.scatter(*zip(*samples))
    >>> ax.set_aspect('equal')
    >>> center = np.array(center)
    >>> normal1 = np.array(normal1)
    >>> normal2 = np.array(normal2)
    >>> vertex1 = center + normal1 * length1 / 2 + normal2 * length2 / 2
    >>> vertex2 = center - normal1 * length1 / 2 + normal2 * length2 / 2
    >>> vertex3 = center - normal1 * length1 / 2 - normal2 * length2 / 2
    >>> vertex4 = center + normal1 * length1 / 2 - normal2 * length2 / 2
    >>> xlim_min = min(vertex1[0], vertex2[0], vertex3[0], vertex4[0])
    >>> xlim_max = max(vertex1[0], vertex2[0], vertex3[0], vertex4[0])
    >>> ylim_min = min(vertex1[1], vertex2[1], vertex3[1], vertex4[1])
    >>> ylim_max = max(vertex1[1], vertex2[1], vertex3[1], vertex4[1])
    >>> graph_limits = geometry.get_limits_of_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max)
    >>> ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    >>> ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
    >>> # create title from n_samples, center, and radius, using f-string
    >>> title = (f'{n_samples} Samples on a Parallelogram with normal vectors ({normal1[0]}, {normal1[1]}) '
    >>>         f'and ({normal2[0]}, {normal2[1]}), center ({center[0]}, {center[1]}), length1 of {length1}, '
    >>>         f'and length2 of {length2}'
    >>>          )
    >>> ax.set_title(title)
    >>> # draw also the parallelogram in red
    >>> vertices = geometry.get_parallelogram_2d_vertices(center, normal1, normal2, length1, length2)
    >>> ax.plot([vertices[0][0], vertices[1][0]], [vertices[0][1], vertices[1][1]], color='r')
    >>> ax.plot([vertices[1][0], vertices[2][0]], [vertices[1][1], vertices[2][1]], color='r')
    >>> ax.plot([vertices[2][0], vertices[3][0]], [vertices[2][1], vertices[3][1]], color='r')
    >>> ax.plot([vertices[3][0], vertices[0][0]], [vertices[3][1], vertices[0][1]], color='r')
    >>> # Draw the X and Y axes in dotted lines
    >>> ax.axhline(0, linestyle='dotted', color='black')
    >>> ax.axvline(0, linestyle='dotted', color='black')
    >>> # Draw the normals at a quarter of their corresponding length
    >>> quarter_length1 = length1 / 8
    >>> quarter_length2 = length2 / 8
    >>> arrow_length1 = quarter_length1 / 2
    >>> arrow_length2 = quarter_length2 / 2
    >>> ax.arrow(center[0], center[1], normal1[0] * quarter_length1, normal1[1] * quarter_length1, 
    >>>         head_width=arrow_length1, head_length=arrow_length2, fc='b', ec='b')
    >>> ax.arrow(center[0], center[1], normal2[0] * quarter_length2, normal2[1] * quarter_length2, 
    >>>         head_width=arrow_length2, head_length=arrow_length1, fc='b', ec='b')
    >>> plt.show()

    |sampling_parallelogram_2d|

    .. |sampling_parallelogram_2d| image:: ../../_static/images/sampling_parallelogram_2d.png
    """
    if seed is not None:
        random.seed(seed)
    normal1 = np.array(normal1)
    normal2 = np.array(normal2) 
    center = np.array(center)
    samples = [__sample_point_parallelogram_2d(normal1, normal2, center, length1, length2) for _ in range(n_samples)]
    return samples

def __sample_point_alligned_parallelogram_2d(min_x: float, max_x: float, min_y: float, max_y: float) -> Tuple[float, float]:
    '''
    Sample a point from a parallelogram with sides parallel to the x and y axes.

    :param min_x: float
        The minimum x coordinate of the parallelogram.
    :param max_x: float
        The maximum x coordinate of the parallelogram.
    :param min_y: float
        The minimum y coordinate of the parallelogram.
    :param max_y: float
        The maximum y coordinate of the parallelogram.
    :return: Tuple[float, float]
        A tuple (x, y) representing the coordinates of the sampled point.

    .. note:: The samples are uniformly distributed on the parallelogram.

    Examples
    --------
    Sample a point from a parallelogram with sides parallel to the x and y axes:

    >>> sample_point_alligned_parallelogram(-1, 1, -1, 1)
    (-0.5, 0.5)
    '''
    x = random.uniform(min_x, max_x)
    y = random.uniform(min_y, max_y)
    return x, y

def sampling_alligned_parallelogram_2d(n_samples: int, min_x: float, max_x: float, min_y: float, max_y: float, 
                                       seed: Optional[int] = None) -> List[Tuple[float, float]]:
    '''
    Sample points from a parallelogram with sides parallel to the x and y axes.

    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param min_x: The minimum x coordinate of the parallelogram.
    :type min_x: float
    :param max_x: The maximum x coordinate of the parallelogram.
    :type max_x: float   
    :param min_y: The minimum y coordinate of the parallelogram.
    :type min_y: float
    :param max_y: The maximum y coordinate of the parallelogram.
    :type max_y: float
    :param seed: The seed to use for the random number generator.
    :type seed: Optional[int]
    :return: A list of tuples (x, y) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float]]
        
    .. note:: The samples are uniformly distributed on the parallelogram.

    Examples
    --------
    Sample 100 points from a parallelogram with sides parallel to the x and y axes:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geometry
    >>> import matplotlib.pyplot as plt

    Define the limits of the parallelogram and the number of points to sample:

    >>> n_samples = 100
    >>> min_x = -3
    >>> max_x = 2
    >>> min_y = -1
    >>> max_y = 5

    Sample the points:

    >>> samples = sampling.sampling_alligned_parallelogram_2d(n_samples=n_samples, min_x=min_x, 
    >>>                                          max_x=max_x, min_y=min_y, max_y=max_y, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [(0.19713399228941864, -0.8499354686639984),
    (-1.6248534081544037, 0.3392644288929365),
    (0.6823560708200622, 3.0601969245374683),
    (1.460897838524227, -0.4783670042235031),
    (-0.8903909015736478, -0.8212166833715779)]

    Plot the samples:

    >>> fig, ax = plt.subplots()
    >>> graph_limits = geometry.get_limits_of_graph_from_limits_of_object(min_x, max_x, min_y, max_y)
    >>> ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    >>> ax.set_ylim(graph_limits[2], graph_limits[3]*1.1)  # Set y-axis limits
    >>> ax.scatter(*zip(*samples))
    >>> ax.set_aspect('equal')
    >>> # create title from n_samples, center, and radius, usign fstring
    >>> title = (f'{n_samples} Samples on an axes alligned Parallelogram with bottom left corner '
    >>>          f'({min_x}, {min_y}) and top right corner ({max_x}, {max_y})'
    >>>     )
    >>> ax.set_title(title)
    >>> # draw also the parallelogram in red
    >>> ax.plot([min_x, max_x], [min_y, min_y], color='r')
    >>> ax.plot([min_x, max_x], [max_y, max_y], color='r')
    >>> ax.plot([min_x, min_x], [min_y, max_y], color='r')
    >>> ax.plot([max_x, max_x], [min_y, max_y], color='r')
    >>> ax.set_xlabel('X')
    >>> ax.set_ylabel('Y')
    >>> # Draw the X and Y axes in dotted lines
    >>> ax.axhline(0, linestyle='dotted', color='black')
    >>> ax.axvline(0, linestyle='dotted', color='black')
    >>> plt.show()

    |sampling_alligned_parallelogram_2d|

    .. |sampling_alligned_parallelogram_2d| image:: ../../_static/images/sampling_alligned_parallelogram_2d.png
    '''
    if seed is not None:
        random.seed(seed)
    samples = [__sample_point_alligned_parallelogram_2d(min_x, max_x, min_y, max_y) for _ in range(n_samples)]
    return samples

def __sample_point_parallelogram_3d(normal1: Tuple[float, float, float], normal2: Tuple[float, float, float], 
                                  center: Tuple[float, float, float], length1: float, length2: float) -> Tuple[float, float, float]:
    '''
    Sample a point from a parallelogram with sides parallel to the vectors normal1 and normal2.

    :param normal1: Tuple[float, float, float]
        The first vector normal to the sides of the parallelogram.
    :param normal2: Tuple[float, float, float]
        The second vector normal to the sides of the parallelogram.
    :param center: Tuple[float, float, float]
        The center of the parallelogram.
    :param length1: float
        The length of the first side of the parallelogram.
    :param length2: float
        The length of the second side of the parallelogram.    
    :return: Tuple[float, float, float]
        A tuple (x, y, z) representing the coordinates of the sampled point.
    '''
    # Generate two random numbers between -length1/2 and length1/2 and -length2/2 and length2/2 respectively
    x = random.uniform(-length1/2, length1/2)
    y = random.uniform(-length2/2, length2/2)
    # Those numbers represent the coordinates of the sampled point in the normal coordinate system
    # Transform the coordinates to the global coordinate system
    projected_point = x * np.array(normal1) + y * np.array(normal2) + np.array(center)
    return projected_point

def __sample_point_parallelepiped_3d(normal1: Tuple[float, float, float], normal2: Tuple[float, float, float], 
                                  normal3: Tuple[float, float, float], center: Tuple[float, float, float], 
                                  length1: float, length2: float, length3: float) -> Tuple[float, float, float]:
    '''
    Sample a point from a parallelepiped with sides parallel to the vectors normal1, normal2 and normal3.

    :param normal1: Tuple[float, float, float]
        The first vector normal to the sides of the parallelepiped.
    :param normal2: Tuple[float, float, float]
        The second vector normal to the sides of the parallelepiped.
    :param normal3: Tuple[float, float, float]
        The third vector normal to the sides of the parallelepiped.
    :param center: Tuple[float, float, float]
        The center of the parallelepiped.
    :param length1: float
        The length of the first side of the parallelepiped.
    :param length2: float
        The length of the second side of the parallelepiped.    
    :param length3: float
        The length of the third side of the parallelepiped.
    :return: Tuple[float, float, float]
        A tuple (x, y, z) representing the coordinates of the sampled point.
    '''
    # Generate two random numbers between -length1/2 and length1/2 and -length2/2 and length2/2 respectively
    x = random.uniform(-length1/2, length1/2)
    y = random.uniform(-length2/2, length2/2)
    z = random.uniform(-length3/2, length3/2)
    # Those numbers represent the coordinates of the sampled point in the normal coordinate system
    # Transform the coordinates to the global coordinate system
    projected_point = x * np.array(normal1) + y * np.array(normal2) + z * np.array(normal3) + np.array(center)
    return projected_point

def sampling_parallelogram_3d(n_samples: int, normal1: Tuple[float, float, float], normal2: Tuple[float, float, float],
                              center: Tuple[float, float, float], length1: float, length2: float, seed: Optional[int] = None) -> List[Tuple[float, float, float]]:
    '''
    Sample n_samples points from a parallelogram
     
    The parallelogram is defined by the vectors normal1 and normal2, the center and the lengths of the sides.
   
    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param normal1: The first vector normal to the sides of the parallelepiped.
    :type normal1: Tuple[float, float, float]
    :param normal2: The second vector normal to the sides of the parallelepiped.
    :type normal2: Tuple[float, float, float]
    :param center: The center of the parallelepiped.
    :type center: Tuple[float, float, float]
    :param length1: The length of the first side of the parallelepiped.
    :type length1: float
    :param length2: The length of the second side of the parallelepiped.
    :type length2: float
    :param seed: The seed to use for the random number generator.
    :type seed: Optional[int]
    :return: A list of tuples (x, y, z) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float, float]]

    .. note:: The samples are uniformly distributed on the parallelogram.

    Examples
    --------
    Sample 100 points from a parallelogram in an arbitrary position and with arbitrary sides:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geometry
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np

    Define the parameters of the parallelogram and the number of points to sample:

    >>> n_samples = 100
    >>> normal1 = (1, 1, 0)
    >>> normal2 = (-2, 1, 1)
    >>> center = (1, 2, 0)
    >>> length1 = 5
    >>> length2 = 4

    Sample the points:

    >>> samples = sampling.sampling_parallelogram_3d(n_samples=n_samples, normal1=normal1, normal2=normal2, 
    ...                                              center=center, length1=length1, length2=length2, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [array([ 5.49704795,  0.79717701, -1.89995698]),
    array([ 2.08946069, -0.23201046, -1.10715705]),
    array([0.76876017, 3.88915402, 0.70679795]),
    array([ 6.26538718,  2.30865317, -1.65224467]),
    array([ 4.37123134, -0.27120202, -1.88081112])]

    Plot the 3D parallelogram and the samples:

    >>> samples = np.array(samples)
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection='3d')
    >>> ax.scatter(samples[:, 0], samples[:, 1], samples[:, 2])
    >>> # create title from n_samples, center, and radius, using fstring
    >>> title = (f'{n_samples} Samples on a 3D Parallelogram with normal vectors {normal1} and {normal2}, '
    >>>         f'center {center}, length1 of {length1} and length2 of {length2}')
    >>> ax.set_title(title)
    >>> # Draw the parallelogram
    >>> vertices = geometry.get_parallelogram_3d_vertices(center, normal1, normal2, length1, length2)
    >>> # Define the edges of the 3d parallelogram
    >>> edges = [(0, 1), (1, 2), (2, 3), (3, 0)]
    >>> # Plot the edges
    >>> for edge in edges:
    >>>     ax.plot([vertices[edge[0]][0], vertices[edge[1]][0]],
    >>>         [vertices[edge[0]][1], vertices[edge[1]][1]],
    >>>         [vertices[edge[0]][2], vertices[edge[1]][2]], color='red')
    >>> # Draw the normals at a quarter of their corresponding length
    >>> quarter_length1 = length1 / 4
    >>> quarter_length2 = length2 / 4
    >>> arrow_length1 = quarter_length1 / 2
    >>> arrow_length2 = quarter_length2 / 2
    >>> ax.quiver(center[0], center[1], center[2], normal1[0], normal1[1], normal1[2], length=arrow_length1, normalize=False, color='red')
    >>> ax.quiver(center[0], center[1], center[2], normal2[0], normal2[1], normal2[2], length=arrow_length2, normalize=False, color='red')
    >>> ax.set_xlabel('X')
    >>> ax.set_ylabel('Y')
    >>> ax.set_zlabel('Z')
    >>> plt.show()

    |sampling_parallelogram_3d|

    .. |sampling_parallelogram_3d| image:: ../../_static/images/sampling_parallelogram_3d.png
    '''
    if seed is not None:
        random.seed(seed)
    normal1 = np.array(normal1)
    normal2 = np.array(normal2)
    center = np.array(center)
    samples = [__sample_point_parallelogram_3d(normal1, normal2, center, length1, length2) for _ in range(n_samples)]
    return samples

def sampling_parallelepiped_3d(n_samples: int, normal1: Tuple[float, float, float], normal2: Tuple[float, float, float],
                              normal3: Tuple[float, float, float], center: Tuple[float, float, float], 
                              length1: float, length2: float, length3: float, seed: Optional[int] = None) -> List[Tuple[float, float, float]]:
    '''
    Sample n_samples points from a parallelepiped
     
    The parallelogram is defined by the vectors normal1, normal2 and normal3, the center and the lengths of the sides.
   
    :param n_samples: The number of samples to generate.
    :type n_samples: int
    :param normal1: The first vector normal to the sides of the parallelepiped.
    :type normal1: Tuple[float, float, float]
    :param normal2: The second vector normal to the sides of the parallelepiped.
    :type normal2: Tuple[float, float, float]
    :param normal3: The third vector normal to the sides of the parallelepiped.
    :type normal3: Tuple[float, float, float]
    :param center: The center of the parallelepiped.
    :type center: Tuple[float, float, float]
    :param length1: The length of the first side of the parallelepiped.
    :type length1: float
    :param length2: The length of the second side of the parallelepiped.
    :type length2: float
    :param length3: The length of the third side of the parallelepiped.
    :type length3: float
    :param seed: The seed to use for the random number generator.
    :type seed: Optional[int]
    :return: A list of tuples (x, y, z) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float, float]]

    .. note:: The samples are uniformly distributed on the parallelepiped.

    Examples
    --------
    Sample 100 points from a parallelepiped in an arbitrary position and with arbitrary sides:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geometry
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np

    Define the parameters of the parallelepiped and the number of points to sample:

    >>> n_samples = 100
    >>> normal1 = (1, 1, 0)
    >>> normal2 = (-2, 1, 1)
    >>> normal3 = (1, -1, 3)
    >>> center = (1, 2, 0)
    >>> length1 = 5
    >>> length2 = 4
    >>> length3 = 3

    Sample the points:

    >>> samples = sampling.sampling_parallelepiped_3d(n_samples=n_samples, normal1=normal1, normal2=normal2, 
    ...                                              normal3=normal3, center=center, length1=length1,
    ...                                              length2=length2, length3=length3, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [array([ 4.82213591,  1.47208906, -3.92469311]),
    array([-1.74561756,  1.03184009,  2.53618024]),
    array([ 6.03115264,  2.54288771, -2.35494829]),
    array([ 0.91594816, -1.49252787, -1.07725051]),
    array([ 1.49163196, -2.02162286,  0.14431054])]
        
    Plot the 3D parallelepiped and the samples:

    >>> samples = np.array(samples)
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection='3d')
    >>> ax.scatter(samples[:, 0], samples[:, 1], samples[:, 2])
    >>> # create title from n_samples, center, and radius, using fstring
    >>> title = (f'{n_samples} Samples on a 3D Parallelepiped with normal vectors {normal1}, {normal2} and {normal3}, '
    >>>         f'center {center}, length1 of {length1}, length2 of {length2} and length3 of {length3}')
    >>> ax.set_title(title)
    >>> # Draw the parallelepiped
    >>> vertices = geometry.get_parallelepiped_3d_vertices(center, normal1, normal2, normal3, length1, length2, length3)
    >>> # Define the edges of the 3d parallelepiped
    >>> edges = [(0, 1), (1, 2), (2, 3), (3, 0), (4, 5), (5, 6),
    >>>         (6, 7), (7, 4), (0, 4), (1, 5), (2, 6), (3, 7)]
    >>> # Plot the edges
    >>> for edge in edges:
    >>>     ax.plot([vertices[edge[0]][0], vertices[edge[1]][0]],
    >>>         [vertices[edge[0]][1], vertices[edge[1]][1]],
    >>>         [vertices[edge[0]][2], vertices[edge[1]][2]], color='red')
    >>> # Draw the normals at a quarter of their corresponding length
    >>> quarter_length1 = length1 / 4
    >>> quarter_length2 = length2 / 4
    >>> arrow_length1 = quarter_length1 / 2
    >>> arrow_length2 = quarter_length2 / 2
    >>> ax.quiver(center[0], center[1], center[2], normal1[0], normal1[1], normal1[2], length=arrow_length1, normalize=False, color='red')
    >>> ax.quiver(center[0], center[1], center[2], normal2[0], normal2[1], normal2[2], length=arrow_length2, normalize=False, color='red')
    >>> ax.quiver(center[0], center[1], center[2], normal3[0], normal3[1], normal3[2], length=arrow_length2, normalize=False, color='red')
    >>> ax.set_xlabel('X')
    >>> ax.set_ylabel('Y')
    >>> ax.set_zlabel('Z')
    >>> plt.show()

    |sampling_parallelepiped_3d|

    .. |sampling_parallelepiped_3d| image:: ../../_static/images/sampling_parallelepiped_3d.png
    '''
    if seed is not None:
        random.seed(seed)
    normal1 = np.array(normal1)
    normal2 = np.array(normal2)
    normal3 = np.array(normal3)
    center = np.array(center)
    samples = [__sample_point_parallelepiped_3d(normal1, normal2, normal3, center, length1, length2, length3) for _ in range(n_samples)]
    return samples

def __sample_point_cuboid(a, b, c, d, h):
    u = random.random()
    v = random.random()
    w = random.random()
    x = a[0] + u * (b[0] - a[0]) + v * (c[0] - a[0]) + w * (d[0] - a[0])
    y = a[1] + u * (b[1] - a[1]) + v * (c[1] - a[1]) + w * (d[1] - a[1])
    z = a[2] + u * (b[2] - a[2]) + v * (c[2] - a[2]) + w * (d[2] - a[2]) + h
    return x, y, z

def sampling_cuboid(n_samples, a, b, c, d, h):
    samples = [__sample_point_cuboid(a, b, c, d, h) for _ in range(n_samples)]
    return samples

def __sample_point_sphere(center:Tuple[float, float, float]=(0, 0, 0), radius:float=1):
    """
    Generate a random point on a sphere with a specified radius and center, using rejection sampling.

    :param center: tuple
        A tuple (x, y, z) representing the center of the circle.
    :param radius: float, optional
        The radius of the circle. Default is 1.
    :return: tuple
        A tuple (x, y, z) representing the coordinates of the sampled point.
    """
    while True:
        # Generate a random point on the circle with radius 1 centered at (0,0)
        x = random.uniform(-1, 1)
        y = random.uniform(-1, 1)
        z = random.uniform(-1, 1)
        
        # Scale the coordinates to the desired radius
        x *= radius
        y *= radius
        z *= radius
        
        # Translate the point to the desired center
        x += center[0]
        y += center[1]
        z += center[2]
        
        # Check if the point is within the circle
        if (x - center[0])**2 + (y - center[1])**2 + (z - center[2])**2 <= radius**2:
            return x, y, z

def sampling_sphere(n_samples:int=1, center:Tuple[float, float, float]=(0,0,0), radius:float=1, seed:Optional[int]=None) \
                        -> List[Tuple[float, float, float]]:
    """
    Generate random samples on a sphere with a specified radius and center.

    :param n_samples: The number of random samples to generate on the circle.
    :type n_samples: int
    :param center: A tuple (x, y, z) representing the center of the circle. Default is (0,0).
    :type center: Tuple[float, float, float]
    :param radius: The radius of the circle. Default is 1.
    :type radius: float
    :param seed: The seed value for the random number generator. Default is None.
    :type seed: Optional[int]
    :return: A list of tuples (x, y, z) representing the coordinates of the sampled points.
    :rtype: List[Tuple[float, float, float]]
        
    .. note:: The samples are uniformly distributed on the sphere.
    
    Examples
    --------
    Sample 100 points from a sphere:

    Required imports:
    
    >>> import mrdja.sampling as sampling
    >>> import mrdja.geometry as geometry
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np

    Define the parameters of the sphere and the number of points to sample:

    >>> n_samples = 100
    >>> center = (2, 3, 1)
    >>> radius = 5

    Sample the points:

    >>> samples = sampling.sampling_sphere(n_samples=n_samples, center=center, radius=radius, seed=42)
    >>> # list the first 5 samples
    >>> samples[:5]
    [(-0.7678926185117723, 5.364712141640124, 2.766994874229113),
    (2.4494148060321668, 0.204406220406967, 1.8926568387590872),
    (3.9813939498822686, 1.4025051651799187, -2.4452050018821847),
    (5.071282732743802, 5.297317866938179, 1.3622809145470074),
    (6.731157639793706, 1.7853437720835348, 1.52040631273227)]
    
    Plot the samples:
    
    >>> samples = np.array(samples)
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection='3d')
    >>> ax.scatter(samples[:, 0], samples[:, 1], samples[:, 2])
    >>> # plot the sphere
    >>> u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
    >>> x = radius*np.cos(u)*np.sin(v) + center[0]
    >>> y = radius*np.sin(u)*np.sin(v) + center[1]
    >>> z = radius*np.cos(v) + center[2]
    >>> ax.plot_wireframe(x, y, z, color="r")
    >>> title = f'{n_samples} Samples on a Sphere with Center ({center[0]}, {center[1]}, {center[2]}) and Radius {radius}'
    >>> ax.set_title(title)
    >>> ax.set_xlabel('X')
    >>> ax.set_ylabel('Y')
    >>> ax.set_zlabel('Z')
    >>> plt.show()
    
    |sampling_sphere|

    .. |sampling_sphere| image:: ../../_static/images/sampling_sphere.png

    """
    # Set the random seed for reproducibility if provided
    if seed is not None:
        __set_random_seed(seed)

    # Generate the specified number of samples on the circle
    samples = [__sample_point_sphere(center, radius) for _ in range(n_samples)]

    return samples

def sampling_np_array_elements(elements:np.ndarray, num_samplings: int = 1, replacement: bool=False, 
                               len_elements: Optional[int]=None, seed: Optional[int]=None) -> np.ndarray:
    """
    Sample elements from a numpy array.

    :param elements: The array of elements to sample from.
    :type elements: np.ndarray
    :param num_samplings: The number of elements to sample. Default is 1.
    :type num_samplings: int
    :param replacement: Whether to sample with replacement or not. Default is False.
    :type replacement: bool
    :param len_elements: The length of the elements array. Default is None.
    :type len_elements: int
    :param seed: The seed value for the random number generator. Default is None.
    :type seed: int
    :return: The sampled elements.
    :rtype: np.ndarray

    Examples
    --------

    >>> import numpy as np
    >>> import mrdja.sampling as sampling
    >>> sampling.sampling_np_array_elements(np.array([1,2,3,4,5]), 3, False, seed=42)
    array([2, 5, 3])
    >>> sampling.sampling_np_array_elements(np.array([(1,2),(3,4),(5,6),(7,8),(9,10)]), 3, False, seed=42)
    array([[ 3,  4],
       [ 9, 10],
       [ 5,  6]])
    """
    if len_elements is None:
        len_elements = len(elements)
    if seed is not None:
        np.random.seed(seed)
    random_elements_indices = np.random.choice(range(len_elements), num_samplings, replace=replacement)
    random_elements = elements[random_elements_indices]
    return random_elements

def sampling_pcd_points(pcd: o3d.geometry.PointCloud, num_points: int = 1, seed: Optional[int] = None):
    """
    Sample points from a point cloud.

    :param pcd: The point cloud to sample from.
    :type pcd: o3d.geometry.PointCloud
    :param num_points: The number of points to sample. Default is 1.
    :type num_points: int
    :param seed: The seed value for the random number generator. Default is None.
    :type seed: int
    :return: The sampled points.
    :rtype: np.ndarray

    Examples
    --------

    >>> import open3d as o3d
    >>> import mrdja.sampling as sampling
    >>> # Create a point cloud from random points sampled from a 3D parallelogram
    >>> n_samples = 1000
    >>> normal1 = (1, 1, 0)
    >>> normal2 = (-2, 1, 1)
    >>> normal3 = (1, -1, 3)
    >>> center = (1, 2, 0)
    >>> length1 = 5
    >>> length2 = 4
    >>> length3 = 3
    >>> samples = sampling.sampling_parallelogram_3d(n_samples=n_samples, normal1=normal1, normal2=normal2, 
                                             normal3=normal3, center=center, length1=length1,
                                             length2=length2, length3=length3, seed=42)
    >>> samples[:5]
    [array([ 4.82213591,  1.47208906, -3.92469311]),
    array([-1.74561756,  1.03184009,  2.53618024]),
    array([ 6.03115264,  2.54288771, -2.35494829]),
    array([ 0.91594816, -1.49252787, -1.07725051]),
    array([ 1.49163196, -2.02162286,  0.14431054])]
    >>> pcd = o3d.geometry.PointCloud()
    >>> pcd.points = o3d.utility.Vector3dVector(samples)
    >>> # paint the point cloud in blue
    >>> pcd.paint_uniform_color([0, 0, 1])
    >>> # Sample 300 random points from the point cloud
    >>> sampled_points = sampling.sampling_pcd_points(pcd, 300, seed=42)
    >>> sampled_points[:5]
    array([[-3.29369778,  2.38822478,  3.96820711],
            [-0.15203634,  3.14597455,  3.11019749],
            [ 2.41599518, -1.37126689, -3.70481201],
            [ 3.02922805, -2.15456602, -0.98563083],
            [ 0.75636732,  5.3772318 , -2.63251376]])
    >>> # create a point cloud from the sampled points
    >>> sampled_pcd = o3d.geometry.PointCloud()
    >>> sampled_pcd.points = o3d.utility.Vector3dVector(sampled_points)
    >>> # paint the sampled points in red
    >>> sampled_pcd.paint_uniform_color([1, 0, 0])
    >>> # visualize the point clouds
    >>> o3d.visualization.draw_geometries([pcd, sampled_pcd])

    |sampling_pcd_points|

    .. |sampling_pcd_points| image:: ../../_static/images/sampling_pcd_points.png
    """

    points = np.asarray(pcd.points)
    random_points = sampling_np_array_elements(elements=points, num_samplings=num_points, seed=seed)
    return random_points

def sampling_np_arrays_from_enumerable(source_list: Union[List, np.ndarray], cardinality_of_np_arrays: int, number_of_np_arrays: int=1, num_source_elems: Optional[int] = None, seed: Optional[int] = None) -> List[np.ndarray]:
    """
    Returns a list with **number_of_np_arrays** numpy arrays of size **cardinality_of_np_arrays** with random elements from a list or numpy array.

    :param source_list: The list or numpy array to sample from.
    :type source_list: Union[list, np.ndarray]
    :param cardinality_of_np_arrays: The cardinality of the numpy arrays to generate.
    :type cardinality_of_np_arrays: int
    :param number_of_np_arrays: The number of numpy arrays to generate. Default is 1.
    :type number_of_np_arrays: int
    :param num_source_elems: The number of elements in the source list or numpy array. Default is None.
    :type num_source_elems: int
    :param seed: The seed value for the random number generator. Default is None.
    :type seed: int
    :return: List of numpy arrays containing the sampled arrays.
    :rtype: List[np.ndarray]

    :Example:

    ::

        >>> import mrdja.sampling as sampling
        >>> import numpy as np
        >>> sampling.sampling_np_arrays_from_enumerable(np.array([1,2,3,4,5,6,7,8,9,10]), 3, 2, seed=42)
        >>> [array([9, 2, 6]), array([1, 8, 3])]
        >>> np.random.seed(42)
        >>> random_3d_points = np.random.rand(100, 3)
        >>> random_np_arrays_of_points = sampling.sampling_np_arrays_from_enumerable(random_3d_points, cardinality_of_np_arrays=3, number_of_np_arrays=4, seed=42)
        >>> random_np_arrays_of_points
        >>> [array([[0.85300946, 0.29444889, 0.38509773],
        >>> [0.72821635, 0.36778313, 0.63230583],
        >>> [0.54873379, 0.6918952 , 0.65196126]]),
        >>> array([[0.32320293, 0.51879062, 0.70301896],
        >>> [0.11986537, 0.33761517, 0.9429097 ],
        >>> [0.18657006, 0.892559  , 0.53934224]]),
        >>> array([[0.14092422, 0.80219698, 0.07455064],
        >>> [0.94045858, 0.95392858, 0.91486439],
        >>> [0.60754485, 0.17052412, 0.06505159]]),
        >>> array([[0.37454012, 0.95071431, 0.73199394],
        >>> [0.59789998, 0.92187424, 0.0884925 ],
        >>> [0.11959425, 0.71324479, 0.76078505]])]

    """
    if num_source_elems is None:
        num_source_elems = len(source_list)
    if seed is not None:
        np.random.seed(seed)
    random_elems_indices = np.random.choice(range(num_source_elems), size= cardinality_of_np_arrays * number_of_np_arrays, replace=False)
    random_elems = np.array(source_list)[random_elems_indices]
    # Split random_elems into sets
    sampled_np_arrays = [random_elems[i * cardinality_of_np_arrays:(i + 1) * cardinality_of_np_arrays] for i in range(number_of_np_arrays)]
    return sampled_np_arrays




import mrdja.ransac.coreransac as coreransac
import mrdja.ransaclpexperiments as experiments
import open3d as o3d
import numpy as np
import random
import pickle as pkl

'''
Line-guided plane detection in pointclouds.
Tomar, por ejemplo, 200 rectas y luego los 40 mejores por parejas. Eso son 780 parejas. Tomar el 5% mejor según el error de ajuste del plano.
Eso son 39 parejas. En total 239 veces que hay que mirar los inliers mirando todos los puntos.
Calcular el mejor plano, tomando RANSAC de open3d con 100.000 iteraciones.
Hacer varios experimentos, tanto de RANSAC con 239 o más iteraciones, como de open3d RANSAC con otros números de iteraciones.
'''

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

import open3d as o3d
import mrdja.ransaclp as ransaclp
import mrdja.ransac.coreransac as coreransac
import numpy as np

dataset = o3d.data.LivingRoomPointClouds()
living_room_paths = dataset.paths
dataset = o3d.data.OfficePointClouds()
office_paths = dataset.paths
ply_files = living_room_paths + office_paths

threshold = 0.02
total_number_files = len(ply_files)

for index, filename in enumerate(ply_files):

    print(f"filename {index} of {total_number_files}: {filename}")

    dict_all_results = {}
    dict_all_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    np_points = np.asarray(pcd.points)
    dict_all_results["number_pcd_points"] = len(np_points)
    dict_all_results["threshold"] = threshold

    for num_iterations in [100, 200, 300, 400, 500, 600]:
        parameters_experiment = experiments.compute_parameters_ransac_line(num_iterations, percentage_chosen_lines = 0.2, percentage_chosen_planes = 0.05)
        number_chosen_lines = parameters_experiment["number_chosen_lines"]
        number_lines_pairs = parameters_experiment["number_lines_pairs"]
        number_chosen_planes = parameters_experiment["number_chosen_planes"]
        total_iterations = parameters_experiment["total_iterations"]
        # filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
        # filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/office_1/office_1.ply"

        dict_standard_RANSAC_results_list = list()
        dict_line_RANSAC_results_list = list()

        for j in range(10):
            ransaclp_data_from_file = ransaclp.get_ransaclp_data_from_filename(filename, ransac_iterations = num_iterations, 
                                                           threshold = threshold, audit_cloud=True)
            ransaclp_number_inliers = ransaclp_data_from_file["number_inliers"]
            ransaclp_plane = ransaclp_data_from_file["plane"]
            
            ransac_plane, inliers = pcd.segment_plane(distance_threshold=threshold,
                                                    ransac_n=3,
                                                    num_iterations=total_iterations)
            ransac_number_inliers = len(inliers)

            dict_standard_RANSAC_results = {"number_inliers": ransac_number_inliers, "plane": ransac_plane, "plane_iterations": total_iterations}
            dict_line_RANSAC_results = {"number_inliers": ransaclp_number_inliers, "plane": ransaclp_plane, "line_iterations": num_iterations}
            dict_line_RANSAC_results_list.append(dict_line_RANSAC_results)
            dict_standard_RANSAC_results_list.append(dict_standard_RANSAC_results)

        dict_all_results["standard_RANSAC_" + str(total_iterations)] = dict_standard_RANSAC_results_list
        dict_all_results["line_RANSAC_" + str(total_iterations)] = dict_line_RANSAC_results_list
        # get the mean of n_inliers_maximum of the elements of the list dict_line_RANSAC_results_list
        list_n_inliers_maximum = [int(dict_line_RANSAC_results["number_inliers"]) for dict_line_RANSAC_results in dict_line_RANSAC_results_list]
        mean_n_inliers_maximum = np.mean(list_n_inliers_maximum)
        dict_all_results["mean_number_inliers_line_RANSAC_" + str(total_iterations)] = mean_n_inliers_maximum
        # get the mean of n_inliers_maximum of the elements of the list dict_standard_RANSAC_results_list
        list_n_inliers_maximum = [int(dict_standard_RANSAC_results["number_inliers"]) for dict_standard_RANSAC_results in dict_standard_RANSAC_results_list]
        mean_n_inliers_maximum = np.mean(list_n_inliers_maximum)
        dict_all_results["mean_number_inliers_standard_RANSAC_" + str(total_iterations)] = mean_n_inliers_maximum

    # save the results as a pickle file in the same folder as the filename file; to do so, just change the extension of the file to pkl
    filename_pkl = filename.replace(".ply", ".pkl")
    # filename is in the form '/root/open3d_data/extract/OfficePointClouds/cloud_bin_51.ply'; convert it to "/tmp/cloud_bin_51.pkl"
    filename_pkl = filename_pkl.split("/")[-1]
    filename_pkl = "/tmp/" + filename_pkl
    with open(filename_pkl, 'wb') as f:
        pkl.dump(dict_all_results, f)





import gradio as gr
import mrdja.sampling as sampling
import mrdja.geometry as geometry
import matplotlib.pyplot as plt
from gradio.components import Number, Image
import matplotlib
import numpy as np
matplotlib.use('Agg')

def plot_alligned_parallelogram(n_samples, min_x, max_x, min_y, max_y):
    samples = sampling.sampling_alligned_parallelogram_2d(n_samples=n_samples, min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y)
    fig, ax = plt.subplots()
    ax.scatter(*zip(*samples))
    ax.set_aspect('equal')
    xlim_min = min_x
    xlim_max = max_x
    ylim_min = min_y
    ylim_max = max_y
    graph_limits = geometry.get_limits_of_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max)
    ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
    # create title from n_samples, center, and radius, using f-string
    ax.set_title(f'{n_samples} Samples on a axes alligned Parallelogram with bottom left corner {min_x}, {min_y} and top right corner {max_x}, {max_y}')
    # draw also the parallelogram in red
    ax.plot([min_x, max_x], [min_y, min_y], color='r')
    ax.plot([min_x, max_x], [max_y, max_y], color='r')
    ax.plot([min_x, min_x], [min_y, max_y], color='r')
    ax.plot([max_x, max_x], [min_y, max_y], color='r')

    # Draw the X and Y axes in dotted lines
    ax.axhline(0, linestyle='dotted', color='black')
    ax.axvline(0, linestyle='dotted', color='black')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    fig.canvas.draw()
    img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close(fig)
    return img

# Define the input and output interfaces for Gradio
inputs = [
    Number(value=100, precision=0, label="Number of Samples"),
    Number(value=0, label="X value of bottom left corner"),
    Number(value=0, label="X value of top right corner"),
    Number(value=1, label="Y value of bottom left corner"),
    Number(value=1, label="Y value of top right corner")
]
outputs = [
    Image(label="Parallelogram Plot", type="pil")
]

# Create the Gradio app
app = gr.Interface(
    fn=plot_alligned_parallelogram, 
    inputs=inputs, 
    outputs=outputs,
    title="Sampling 2D Alligned Parallelograms",
    description="Generate random samples on a 2D alligned parallelogram",
    allow_flagging="never",
)

# Run the app
app.launch()



import mrdja.ransac.coreransac as coreransac
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np
import random
import pickle as pkl

'''
Line-guided plane detection in pointclouds.
Tomar, por ejemplo, 200 rectas y luego los 40 mejores por parejas. Eso son 780 parejas. Tomar el 5% mejor según el error de ajuste del plano.
Eso son 39 parejas. En total 239 veces que hay que mirar los inliers mirando todos los puntos.
Calcular el mejor plano, tomando RANSAC de open3d con 100.000 iteraciones.
Hacer varios experimentos, tanto de RANSAC con 239 o más iteraciones, como de open3d RANSAC con otros números de iteraciones.
'''

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

# database_path = "/home/scpmaotj/Stanford3dDataset_v1.2/"
# get a list with all the ply files under database_path
import glob
import os

database_path = "/home/scpmaotj/Stanford3dDataset_v1.2/"
ply_files = glob.glob(database_path + "/**/*.ply", recursive=True)

print(ply_files)

threshold = 0.02
# seed = 42
# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"

for filename in ply_files:

    print("filename", filename)

    dict_all_results = {}
    dict_all_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    np_points = np.asarray(pcd.points)

    plane_model, inliers = pcd.segment_plane(distance_threshold=threshold,
                                            ransac_n=3,
                                            num_iterations=100000)
    [a, b, c, d] = plane_model
    print("Plane equation: {:.2f}x + {:.2f}y + {:.2f}z + {:.2f} = 0".format(a, b, c, d))
    inlier_cloud = pcd.select_by_index(inliers)
    inlier_cloud.paint_uniform_color([1, 0, 0])
    # o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers  " + str(len(inliers)) + " inliers")
    print("Number of inliers: " + str(len(inliers)))

    dict_baseline_results = {"n_points_pcd": len(np_points), "n_inliers_maximum": str(len(inliers)), "best_plane": plane_model, "threshold": threshold}
    dict_all_results["standard_RANSAC_100000"] = dict_baseline_results

    # save the results as a pickle file in the same folder as the filename file; to do so, just change the extension of the file to pkl
    filename_pkl = filename.replace(".ply", "_100000.pkl")
    with open(filename_pkl, 'wb') as f:
        pkl.dump(dict_all_results, f)


  


import mrdja.sampling as sampling
import mrdja.geometry as geom
import mrdja.drawing as drawing
import mrdja.matplot3d as plt3d
import matplotlib.pyplot as plt
import numpy as np

n_samples = 5
normal1 = (1, 0, 0)
normal2 = (0, 1, 0)
center = (.5, .5, 0)
length1 = 1
length2 = 1

points_plane = sampling.sampling_parallelogram_3d(n_samples=n_samples, normal1=normal1, normal2=normal2, 
                                          center=center, length1=length1, length2=length2, seed=43)
points_plane = np.array(points_plane)

points_plane = np.array([[0.8, 0.2, 0], [0.1, 0.8, 0], [0.2, 0.25, 0], [0.6, 0.75, 0], [0.5, 0.5, 0], [0.2, 0.95, 0]])
points_outside_plane = np.array([[0, 0.8, 0.6], [0.6, 1, 0.8], [0.9, 1, 0.6], [0.9, 1, 0.2]])

plane = np.array([0, 0, 1, 0])
cube_min = np.array([0, 0, 0])
cube_max = np.array([1, 1, 1])

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')

color_plane = "green"
color_good_points = "green"
color_bad_points = "red"
color_good_lines = "green"
color_bad_lines = "red"
alpha = 0.1
polygon_points = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
plt3d.draw_polygon(polygon_points, color=color_plane, alpha=alpha, ax=ax)
plt3d.draw_points(points_plane, color=color_good_points, style="o", ax=ax)
plt3d.draw_points(points_outside_plane, color=color_bad_points, style="o", ax=ax)

# save the figure
plt.savefig("plane.png")

good_segment_1 = np.array([points_plane[0], points_plane[2]])
good_segment_2 = np.array([points_plane[4], points_plane[1]])

plt3d.draw_segment(good_segment_1, color=color_good_lines, ax=ax)
plt3d.draw_segment(good_segment_2, color=color_good_lines, ax=ax)

bad_segment_1 = np.array([points_plane[3], points_outside_plane[1]])
bad_segment_2 = np.array([points_plane[5], points_outside_plane[0]])
bad_segment_3 = np.array([points_outside_plane[2], points_outside_plane[3]])

plt3d.draw_segment(bad_segment_1, color=color_bad_lines, ax=ax)
plt3d.draw_segment(bad_segment_2, color=color_bad_lines, ax=ax)
plt3d.draw_segment(bad_segment_3, color=color_bad_lines, ax=ax)

# save the figure
plt.savefig("lines.png")

bad_plane = geom.get_best_plane_from_points_from_two_segments(bad_segment_1, bad_segment_2)[0]
perpendicular1, perpendicular2 = geom.get_two_perpendicular_unit_vectors_in_plane(plane)
points_of_bad_segments = np.vstack((bad_segment_1, bad_segment_2))
bad_centroid = geom.get_centroid_of_points(points_of_bad_segments)
bad_polygon = geom.get_a_polygon_from_plane_equation_and_point(bad_plane, bad_centroid, scale = 0.5)

plt3d.draw_polygon(bad_polygon, color="red", alpha=alpha, ax=ax)

# save the figure
plt.savefig("bad_polygon.png")

plt.show()


# visualize a point cloud in matplotlib

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import open3d as o3d
import mrdja.ransac.coreransac as coreransac

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
pcd_all = o3d.io.read_point_cloud(filename)
np_points = np.asarray(pcd_all.points)
max_x = np.max(np_points[:, 0])
min_x = np.min(np_points[:, 0])
max_y = np.max(np_points[:, 1])
min_y = np.min(np_points[:, 1])
max_z = np.max(np_points[:, 2])
min_z = np.min(np_points[:, 2])
number_pcd_points = len(pcd_all.points)
np_all_points = np.asarray(pcd_all.points)
print("number_pcd_points: ", number_pcd_points)
# downsample the point cloud to make it easier to visualize
pcd = pcd_all.voxel_down_sample(voxel_size=0.1)
number_pcd_points = len(pcd.points)
print("number_pcd_points: ", number_pcd_points)
np_points = np.asarray(pcd.points)
np_colors = np.asarray(pcd.colors)
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')


# Create a meshgrid for plotting the plane defined by A and B
x = np.linspace(min_x-1, max_x+1, 10)
y = np.linspace(min_y-1, max_y+1, 10)
X, Y = np.meshgrid(x, y)

np.random.seed(42)
for i in range(0, 30):
# perform a ransac iteration to find a plane
    dict_results = coreransac.get_ransac_iteration_results(np_all_points, threshold=0.05)
    current_plane = dict_results["current_plane"]
    print("current_plane: ", current_plane)
    A, B, C, D = current_plane
    ax.plot_surface(X, Y, (-(A * X + B * Y + D) / C), alpha=0.01)
ax.scatter(np_points[:, 0], np_points[:, 1], np_points[:, 2], c=np_colors)
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
# set axis limits to make the plot look better. The limits are based on the point cloud
ax.set_xlim(min_x-1, max_x+1)
ax.set_ylim(min_y-1, max_y+1)
ax.set_zlim(min_z-1, max_z+1)
plt.show()


import mrdja.ransac.coreransac as coreransac
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

def get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed):
    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    '''
    audit_before_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_before_sanitizing"] = audit_before_sanitizing
    pcd = pointcloud.pointcloud_sanitize(pcd)
    audit_after_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_after_sanitizing"] = audit_after_sanitizing
    '''
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    dict_full_results["ransac_iterations_results"] = []

    np.random.seed(seed)
    max_number_inliers = 0
    best_iteration_results = None
    for _ in range(ransac_iterations):
        dict_iteration_results = coreransac.get_ransac_iteration_results(np_points, threshold, number_pcd_points)
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

ransac_iterations = 1000
threshold = 0.02
seed = 42

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
RANSAC_data_from_file = get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed)

print(RANSAC_data_from_file)

indices_inliers = RANSAC_data_from_file["ransac_best_iteration_results"]["indices_inliers"]
# create a point cloud with the inliers
pcd = o3d.io.read_point_cloud(filename)
inlier_cloud = pcd.select_by_index(indices_inliers)
o3d.visualization.draw_geometries([inlier_cloud])
# paint the inliers red
inlier_cloud.paint_uniform_color([1, 0, 0])
o3d.visualization.draw_geometries([pcd, inlier_cloud])

# save RANSAC data to file as pickle
import pickle
with open('RANSAC_data_from_file.pickle', 'wb') as handle:
    pickle.dump(RANSAC_data_from_file, handle, protocol=pickle.HIGHEST_PROTOCOL)

# create a np array with all the points
list_iterations_results = RANSAC_data_from_file["ransac_iterations_results"]
all_points = np.empty((0, 3))
for iteration_results in list_iterations_results:
    current_points = iteration_results["current_random_points"]
    print(current_points.shape)
    print(all_points.shape)
    all_points = np.append(all_points, current_points, axis=0)
print(all_points.shape)

# create a np array with all the planes
all_planes = np.empty((0, 4))
for iteration_results in list_iterations_results:
    current_plane = iteration_results["current_plane"]
    # convert current_plane shape from (4,) to (1, 4)
    current_plane = np.expand_dims(current_plane, axis=0)
    print(current_plane)
    print(current_plane.shape)
    print(all_planes.shape)
    all_planes = np.append(all_planes, current_plane, axis=0)
print(all_planes.shape)

# create a np array of dimensions (all_planes.shape[0], all_points.shape[0]) with the distances of each plane to all the points
all_distances = np.empty((all_planes.shape[0], all_points.shape[0]))
for i in range(all_planes.shape[0]):
    current_plane = all_planes[i]
    print(current_plane)
    print(current_plane.shape)
    current_distances = geometry.get_distance_from_points_to_plane(all_points, current_plane)
    print(current_distances.shape)
    all_distances[i] = current_distances

import numpy as np
from sklearn.metrics import silhouette_score
import matplotlib.pyplot as plt
from sklearn.cluster import AgglomerativeClustering

# Perform hierarchical/agglomerative clustering and find the best clustering using silhouette score
best_score = -1
best_n_clusters = -1
best_labels = None
for n_clusters in range(2, 10):
    clusterer = AgglomerativeClustering(n_clusters=n_clusters)
    cluster_labels = clusterer.fit_predict(all_distances)
    silhouette_avg = silhouette_score(all_distances, cluster_labels)
    print("For n_clusters =", n_clusters, "The average silhouette_score is :", silhouette_avg)
    if silhouette_avg > best_score:
        best_score = silhouette_avg
        best_n_clusters = n_clusters
        best_labels = cluster_labels

# project the data points to a 2D space using PCA and color the points according to cluster labels
from sklearn.decomposition import PCA
pca = PCA(n_components=2).fit(all_distances)
pca_2d = pca.transform(all_distances)
plt.scatter(pca_2d[:, 0], pca_2d[:, 1], c=best_labels)
plt.show()

# project the data points to a 3D space using PCA and color the points according to cluster labels
from sklearn.decomposition import PCA
pca = PCA(n_components=3).fit(all_distances)
pca_3d = pca.transform(all_distances)
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(pca_3d[:, 0], pca_3d[:, 1], pca_3d[:, 2], c=best_labels)
plt.show()

# count the number of elements in each cluster
unique, counts = np.unique(best_labels, return_counts=True)

# get the index of an element from the most populated cluster
most_populated_cluster = np.argmax(counts)
# get the first index from best_labels that has the value most_populated_cluster
indices_most_populated_cluster = np.where(best_labels == most_populated_cluster)[0]
# get the planes that correspond to the indices of the most populated cluster
planes_most_populated_cluster = all_planes[indices_most_populated_cluster]
# get the points that correspond to the indices of the most populated cluster, taking into account that 
# the points are in the same order as the planes, and there are three of them for each plane
index_first_point_most_populated_cluster = indices_most_populated_cluster * 3
# each element of points_most_populated_cluster is a np array of shape (3,) with the coordinates of the three points
# that define the plane
points_most_populated_cluster = np.empty((0, 3, 3))
for i in range(len(index_first_point_most_populated_cluster)):
    current_index_first_point = index_first_point_most_populated_cluster[i]
    current_point_first = all_points[current_index_first_point]
    current_point_second = all_points[current_index_first_point+1]
    current_point_third = all_points[current_index_first_point+2]
    current_points = np.array([[current_point_first, current_point_second, current_point_third]])
    points_most_populated_cluster = np.append(points_most_populated_cluster, current_points, axis=0)


# get the geometries of all the planes using drawing.draw_plane_as_lines_open3d(*closest_plane, line_color=[0, 1, 1])
planes_most_populated_cluster_geometries = []
for plane in planes_most_populated_cluster:
    plane_geometry = drawing.draw_plane_as_lines_open3d(*plane, line_color=[0, 1, 1])
    planes_most_populated_cluster_geometries.append(plane_geometry)
# draw the geometries of the planes along with the point cloud
o3d.visualization.draw_geometries([pcd, *planes_most_populated_cluster_geometries])
plane_geometry = drawing.draw_plane_as_lines_open3d(*planes_most_populated_cluster[0], line_color=[0, 1, 1])
o3d.visualization.draw_geometries([pcd, plane_geometry])




import pandas as pd
import numpy as np
from scipy.stats import friedmanchisquare
from scikit_posthocs import posthoc_nemenyi

def friedman_nemenyi_test(dataframe):
    # Make a copy of the dataframe to avoid modifying the original one
    df = dataframe.copy()
    
    # Replace '_' (not preceded by a backslash) with '\_' in column names for LaTeX compatibility
    df.columns = df.columns.str.replace(r'(?<!\\)_', r'\_')
    
    # Compute mean values for each algorithm
    means = df.mean()

    means_2 = means.copy()
    means_2.reset_index(drop=True, inplace=True)
    means_2.index = means_2.index.astype(str)

    # Perform Friedman test
    ranks = df.rank(axis=1, method='average', ascending=False)
    friedman_result = friedmanchisquare(*ranks.values.T)

    # Compute mean ranks
    mean_ranks = ranks.mean()

    # Transpose the DataFrame to get the algorithms as rows and datasets as columns
    df = df.T

    # Perform Nemenyi posthoc test
    nemenyi_result = posthoc_nemenyi(df.values)
    nemenyi_result.columns = df.index
    nemenyi_result.index = df.index

    # Create a DataFrame of the Nemenyi results
    nemenyi_df = pd.DataFrame(nemenyi_result)

    # Create a custom function to convert the DataFrame to a LaTeX table
    def to_latex_custom(df, means):
        header = ' & ' + ' & '.join(df.columns) + ' \\\\ \\hline'
        rows = []
        for row_name, row in df.iterrows():
            formatted_values = []
            for col_name, val in row.iteritems():
                if val < 0.05:
                    direction = '+' if means[row_name] > means[col_name] else '-'
                    formatted_values.append(f"\\bf{{{val:.3f} {direction}}}")
                else:
                    formatted_values.append(f"{val:.3f}")
            rows.append(f"{row_name} & " + ' & '.join(formatted_values) + ' \\\\ \\hline')
        return '\\begin{tabular}{|' + 'c|' * (df.shape[1] + 1) + '}\n' + header + '\n' + '\n'.join(rows) + '\n\\end{tabular}'

    # Convert the DataFrame to a LaTeX table
    nemenyi_table = to_latex_custom(nemenyi_df, means)

    nemenyi_df_2 = nemenyi_df.copy()
    nemenyi_df_2.reset_index(drop=True, inplace=True)
    nemenyi_df_2.columns = nemenyi_df_2.index
    # change the column names to strings
    nemenyi_df_2.columns = nemenyi_df_2.columns.astype(str)

    nemenyi_table_2 = to_latex_custom(nemenyi_df_2, means_2)

    results = {
        'friedman_result': friedman_result,
        'mean_ranks': mean_ranks,
        'nemenyi_result': nemenyi_result,
        'nemenyi_df': nemenyi_df,
        'nemenyi_table': nemenyi_table,
        'nemenyi_df_2': nemenyi_df_2,
        'nemenyi_table_2': nemenyi_table_2
    }

    return results











import numpy as np
import open3d as o3d
from time import time
import math
from typing import Tuple, List
from numba import cuda

@cuda.jit
def get_how_many_below_threshold_kernel(points_x: np.ndarray, points_y: np.ndarray, points_z: np.ndarray,
                                         a: float, b: float, c: float, d: float, a1: float, b1: float, c1: float, d1: float,
                                         optimized_threshold: float, point_indices: np.ndarray, point_indices_1: np.ndarray) -> None:
    i = cuda.grid(1)
    if i < points_x.shape[0]:
        dist = math.fabs(a * points_x[i] + b * points_y[i] + c * points_z[i] + d)
        # if dist <= optimized_threshold:
        #     point_indices[i] = 1
        point_indices[i] = int(dist <= optimized_threshold) * 2 - 1
        dist1 = math.fabs(a1 * points_x[i] + b1 * points_y[i] + c1 * points_z[i] + d1)
        # if dist1 <= optimized_threshold:
        #    point_indices_1[i] = 1
        point_indices_1[i] = int(dist1 <= optimized_threshold) * 2 - 1

def get_how_many_below_threshold_between_plane_and_points_and_their_indices_cuda(
    num_points: int, d_points_x: np.ndarray, d_points_y: np.ndarray, d_points_z: np.ndarray, 
    plane: Tuple[float, float, float, float], plane1: Tuple[float, float, float, float], threshold: float) -> Tuple[int, List[int]]:
    # t1 = time()
    a = plane[0]
    b = plane[1]
    c = plane[2]
    d = plane[3]
    a1 = plane1[0]
    b1 = plane1[1]
    c1 = plane1[2]
    d1 = plane1[3]
    optimized_threshold = threshold * math.sqrt(a * a + b * b + c * c)
    point_indices = np.empty(num_points, dtype=np.int64)
    point_indices_1 = np.empty(num_points, dtype=np.int64)
    # fill point_indices with -1
    point_indices[:] = -1
    threadsperblock = 512
    blockspergrid = math.ceil(num_points / threadsperblock)
    # point_indices = cuda.device_array(point_indices.shape, dtype=point_indices.dtype)
    d_point_indices = cuda.to_device(point_indices)
    d_point_indices_1 = cuda.to_device(point_indices_1)
    # t2 = time()
    get_how_many_below_threshold_kernel[blockspergrid, threadsperblock](d_points_x, d_points_y, d_points_z, a, b, c, d, a1, b1, c1, d1, optimized_threshold, d_point_indices, d_point_indices_1)
    # t3 = time()
    point_indices = d_point_indices.copy_to_host()
    # get the count of point_indices that are not -1
    count = np.count_nonzero(point_indices != -1)
    # get the indices of the points that are not -1
    new_indices = np.where(point_indices != -1)
    new_indices = new_indices[0].tolist()
    '''
    t4 = time()
    print(f't2 - t1: {(t2-t1):.4f}s')
    print(f't3 - t2: {(t3-t2):.4f}s')
    print(f't4 - t3: {(t4-t3):.4f}s')
    print(f't4 - t1: {(t4-t1):.4f}s')
    '''
    return count, new_indices

filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"

pcd = o3d.io.read_point_cloud(filename)

np_points = np.asarray(pcd.points)

# Extract x, y, z coordinates from np_points
points_x = np_points[:, 0]
points_y = np_points[:, 1]
points_z = np_points[:, 2]

# Convert points_x, points_y, and points_z to contiguous arrays
d_points_x = np.ascontiguousarray(points_x)
d_points_y = np.ascontiguousarray(points_y)
d_points_z = np.ascontiguousarray(points_z)

num_points = np_points.shape[0]

t1 = time()
for i in range(500):
    a, b, c, d = np.random.rand(4)
    a1, b1, c1, d1 = np.random.rand(4)
    new_plane = (a, b, c, d)
    new_plane1 = (a1, b1, c1, d1)
    threshold_pcd = 0.1

    how_many, inliers = get_how_many_below_threshold_between_plane_and_points_and_their_indices_cuda(num_points, 
                                                                                                    d_points_x,
                                                                                                    d_points_y,
                                                                                                    d_points_z,
                                                                                                    new_plane, new_plane1, threshold_pcd)
t2 =time()
print(f't2 - t1: {(t2-t1):.4f}s')


'''
RANSAC functions

'''

import open3d as o3d
import mrdja.ransac.coreransac as coreransac
import mrdja.ransac.coreransaccuda as coreransaccuda
import mrdja.geometry as geometry
import random
import numpy as np
from typing import Tuple, Dict, Union

def o3d_plane_finder(pcd: o3d.geometry.PointCloud, distance_threshold: float, 
                          num_iterations: int) -> Dict[str, Union[np.ndarray, o3d.geometry.PointCloud]]:
    """
    Uses Open3D to find the best plane that fits a point cloud using RANSAC algorithm.

    :param pcd: Point cloud to fit a plane.
    :type pcd: o3d.geometry.PointCloud
    :param distance_threshold: Maximum distance a point can have to be considered in the plane.
    :type distance_threshold: float
    :param num_iterations: Number of iterations to find the plane.
    :type num_iterations: int
    :return: Parameters of the best plane model and two point clouds containing the inliers and outliers, respectively.
    :rtype: Tuple[np.ndarray, o3d.geometry.PointCloud, o3d.geometry.PointCloud]
    """
    plane_model, inliers = pcd.segment_plane(distance_threshold=distance_threshold, ransac_n=3, num_iterations=num_iterations)
    # [a, b, c, d] = plane_model
    # print("Plane equation: {}x + {}y + {}z + {} = 0".format(a, b, c, d))
    inlier_cloud = pcd.select_by_index(inliers)
    outlier_cloud = pcd.select_by_index(inliers, invert=True)
    return {"plane_model": plane_model, "inlier_cloud": inlier_cloud, "outlier_cloud": outlier_cloud}

def ransac_from_scratch_plane_finder(pcd: o3d.geometry.PointCloud, 
                                      distance_threshold: float, 
                                      num_iterations: int) -> Dict[str, Union[np.ndarray, o3d.geometry.PointCloud]]:
    """
    Computes the best plane that fits a collection of points and the indices of the inliers.
    
    :param pcd: The point cloud data.
    :type pcd: o3d.geometry.PointCloud
    :param distance_threshold: Maximum distance to the plane.
    :type distance_threshold: float
    :param num_iterations: Number of iterations to compute the best plane.
    :type num_iterations: int
    :return: Best plane parameters and the indices of the points that fit the plane.
    :rtype: Tuple[np.ndarray, o3d.geometry.PointCloud, o3d.geometry.PointCloud]
    """
    num_points = len(pcd.points)
    pcd_points = np.asarray(pcd.points, dtype="float32") 
    dict_results = coreransac.get_ransac_results(pcd_points, num_points, distance_threshold, num_iterations)
    plane_model = dict_results["best_plane"]
    best_inlier_indices = dict_results["indices_inliers"]   
    inlier_cloud = pcd.select_by_index(best_inlier_indices)
    outlier_cloud = pcd.select_by_index(best_inlier_indices, invert=True)
            
    return {"plane_model": plane_model, "inlier_cloud": inlier_cloud, "outlier_cloud": outlier_cloud}

def cuda_plane_finder(pcd: o3d.geometry.PointCloud, 
                      distance_threshold: np.float32, 
                      num_iterations: np.int64) -> Dict[str, Union[np.ndarray, o3d.geometry.PointCloud]]:
    """
    Computes the best plane that fits a point cloud and the indices of the inliers using GPU acceleration.
    
    :param pcd: The point cloud.
    :type pcd: o3d.geometry.PointCloud
    :param distance_threshold: Maximum distance to the plane.
    :type distance_threshold: np.float32
    :param num_iterations: Number of iterations to compute the best plane.
    :type num_iterations: np.int64
    :return: The best plane parameters, the inliers point cloud and the outliers point cloud.
    :rtype: Tuple[np.ndarray, o3d.geometry.PointCloud, o3d.geometry.PointCloud]
    """
    num_points = len(pcd.points)
    if num_points == 0:
        raise ValueError("Input point cloud is empty")   
    pcd_points = np.asarray(pcd.points, dtype="float32") 
    dict_results = coreransaccuda.get_ransac_results_cuda(pcd_points, num_points, distance_threshold, num_iterations)
    plane_model = dict_results["best_plane"]
    best_inlier_indices = dict_results["indices_inliers"]   
    inlier_cloud = pcd.select_by_index(best_inlier_indices)
    outlier_cloud = pcd.select_by_index(best_inlier_indices, invert=True)
            
    return {"plane_model": plane_model, "inlier_cloud": inlier_cloud, "outlier_cloud": outlier_cloud}

def find_multiple_planes(pcd: o3d.geometry.PointCloud, distance_threshold: float, num_iterations: int, num_planes: int, finder_function) -> Dict[str, Union[np.ndarray, o3d.geometry.PointCloud]]:
    planes = []
    for _ in range(num_planes):
        if len(pcd.points) < 3:
            break
        d = finder_function(pcd, distance_threshold, num_iterations)
        planes.append(d)
        pcd = d["outlier_cloud"]
    return planes

def color_multiple_planes(pcd: o3d.geometry.PointCloud, distance_threshold: float, num_iterations: int, num_planes: int, finder_function) -> o3d.geometry.PointCloud:
    planes = find_multiple_planes(pcd, distance_threshold, num_iterations, num_planes, finder_function)
    colors = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0], [1, 0, 1], [0, 1, 1]]
    colored_pcd = o3d.geometry.PointCloud()
    for i, plane in enumerate(planes):
        color = colors[i % len(colors)]
        plane["inlier_cloud"].paint_uniform_color(color)
        colored_pcd += plane["inlier_cloud"]
    colored_pcd += planes[-1]["outlier_cloud"]
    return colored_pcd

def get_plane_distances_to_points(plane: np.ndarray, points: np.ndarray) -> np.ndarray:
    """
    Computes the distances from a plane to a collection of points.

    :param plane: Plane parameters.
    :type plane: np.ndarray
    :param points: Collection of points.
    :type points: np.ndarray
    :return: Distances from the plane to the points.
    :rtype: np.ndarray

    :Example:

    ::

        >>> import mrdja.ransac as ran
        >>> import numpy as np
        >>> points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]], dtype="float32")
        >>> plane = np.array([0, 0, 1, 0], dtype="float32")
        >>> ran.get_plane_distances_to_points(plane, points)
        array([ 0.,  0.,  0.], dtype=float32)
    """
    '''
    if it is a list convert it to numpy array, and check if it is empty
    '''
    if isinstance(points, list):
        points = np.asarray(points)
    if len(points) == 0:
        return np.array([], dtype="float32")
    return np.abs(np.dot(points, plane[:3]) + plane[3]) / np.linalg.norm(plane[:3])

def generate_k_planes_from_k_sets_of_3_random_points_of_pointcloud_points_and_compute_the_distance_from_the_planes_to_all_the_points(pcd: o3d.geometry.PointCloud, k: int) -> np.ndarray:
    """
    Generates k planes from k sets of 3 random points of point cloud points and computes the distance from the planes to all the points.

    :param pcd: Point cloud data.
    :type pcd: o3d.geometry.PointCloud
    :param k: Number of planes to generate.
    :type k: int
    :return: Distances from the planes to the points.
    :rtype: np.ndarray

    :Example:

    ::

        >>> import mrdja.ransac as ran
        >>> import numpy as np
        >>> import open3d as o3d
        >>> pcd = o3d.geometry.PointCloud()
        >>> pcd.points = o3d.utility.Vector3dVector(np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 2, 3], [5, 6, 7]], dtype="float32"))
        >>> ran.generate_k_planes_from_k_sets_of_3_random_points_of_pointcloud_points_and_compute_the_distance_from_the_planes_to_all_the_points(pcd, 3)
    """
    num_points = len(pcd.points)
    pcd_points = np.asarray(pcd.points, dtype="float32")
    points_to_compute_distance = np.empty((0, 3), dtype="float32")
    all_planes = []
    dict_distances = {}
    for _ in range(k):
        current_points = coreransac.get_np_array_of_three_random_points_from_np_array_of_points(pcd_points, num_points)
        current_plane = geometry.get_plane_from_list_of_three_points(current_points)
        get_plane_distances_to_points(current_plane, points_to_compute_distance)
        dict_distances[tuple(current_plane)] = get_plane_distances_to_points(current_plane, points_to_compute_distance)
        all_planes.append(current_plane)
        for plane in all_planes:
            previous_distances = dict_distances[tuple(plane)]
            current_distances = get_plane_distances_to_points(plane, current_points)
            all_distances = np.concatenate((previous_distances, current_distances))
            # convert plane to tuple to use it as key in dictionary
            dict_distances[tuple(plane)] = all_distances
        points_to_compute_distance = np.concatenate((points_to_compute_distance, current_points))
    return dict_distances



import gradio as gr
import mrdja.sampling as sampling
import mrdja.geometry as geometry
import matplotlib.pyplot as plt
from gradio.components import Number, Image
import matplotlib
import numpy as np
matplotlib.use('Agg')

def plot_parallelogram(n_samples, normal1_x, normal1_y, normal2_x, normal2_y, center_x, center_y, length1, length2):
    samples = sampling.sampling_parallelogram_2d(n_samples=n_samples, normal1=(normal1_x, normal1_y), normal2=(normal2_x, normal2_y), center=(center_x, center_y), length1=length1, length2=length2)
    fig, ax = plt.subplots()
    ax.scatter(*zip(*samples))
    ax.set_aspect('equal')
    center = np.array([center_x, center_y])
    normal1 = np.array([normal1_x, normal1_y])
    normal2 = np.array([normal2_x, normal2_y])
    vertex1 = center + normal1 * length1 / 2 + normal2 * length2 / 2
    vertex2 = center - normal1 * length1 / 2 + normal2 * length2 / 2
    vertex3 = center - normal1 * length1 / 2 - normal2 * length2 / 2
    vertex4 = center + normal1 * length1 / 2 - normal2 * length2 / 2
    xlim_min = min(vertex1[0], vertex2[0], vertex3[0], vertex4[0])
    xlim_max = max(vertex1[0], vertex2[0], vertex3[0], vertex4[0])
    ylim_min = min(vertex1[1], vertex2[1], vertex3[1], vertex4[1])
    ylim_max = max(vertex1[1], vertex2[1], vertex3[1], vertex4[1])
    graph_limits = geometry.get_limits_of_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max)
    ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
    # create title from n_samples, center, and radius, using f-string
    ax.set_title(f'{n_samples} Samples on a Parallelogram with normal vectors {normal1} and {normal2}, center {center}, length1 {length1}, and length2 {length2}')
    # draw also the parallelogram in red
    vertices = geometry.get_parallelogram_2d_vertices(center, normal1, normal2, length1, length2)
    ax.plot([vertices[0][0], vertices[1][0]], [vertices[0][1], vertices[1][1]], color='r')
    ax.plot([vertices[1][0], vertices[2][0]], [vertices[1][1], vertices[2][1]], color='r')
    ax.plot([vertices[2][0], vertices[3][0]], [vertices[2][1], vertices[3][1]], color='r')
    ax.plot([vertices[3][0], vertices[0][0]], [vertices[3][1], vertices[0][1]], color='r')

    # Draw the X and Y axes in dotted lines
    ax.axhline(0, linestyle='dotted', color='black')
    ax.axvline(0, linestyle='dotted', color='black')

    # Draw the normals at a quarter of their corresponding length
    quarter_length1 = length1 / 8
    quarter_length2 = length2 / 8
    arrow_length1 = quarter_length1 / 2
    arrow_length2 = quarter_length2 / 2
    ax.arrow(center_x, center_y, normal1_x * quarter_length1, normal1_y * quarter_length1, head_width=arrow_length1, head_length=arrow_length2, fc='b', ec='b')
    ax.arrow(center_x, center_y, normal2_x * quarter_length2, normal2_y * quarter_length2, head_width=arrow_length2, head_length=arrow_length1, fc='b', ec='b')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    fig.canvas.draw()
    img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close(fig)
    return img

# Define the input and output interfaces for Gradio
inputs = [
    Number(value=100, precision=0, label="Number of Samples"),
    Number(value=0, label="X value of the first normal vector"),
    Number(value=1, label="Y value of the first normal vector"),
    Number(value=1, label="X value of the second normal vector"),
    Number(value=0, label="Y value of the second normal vector"),
    Number(value=0, label="X value of the center of the parallelogram"),
    Number(value=0, label="Y value of the center of the parallelogram"),
    Number(value=1, label="Length of the first side of the parallelogram"),
    Number(value=1, label="Length of the second side of the parallelogram")
]
outputs = [
    Image(label="Parallelogram Plot", type="pil")
]

# Create the Gradio app
app = gr.Interface(
    fn=plot_parallelogram, 
    inputs=inputs, 
    outputs=outputs,
    title="Sampling 2D Parallelograms",
    description="Generate random samples on a 2D parallelogram",
    allow_flagging="never",
)

# Run the app
app.launch()



import mrdja.sampling as sampling
import matplotlib.pyplot as plt

n_samples = 100
min_x = -3
max_x = 2
min_y = -1
max_y = 5
samples = sampling.sampling_alligned_parallelogram_2d(n_samples, min_x, max_x, min_y, max_y)
fig, ax = plt.subplots()
ax.scatter(*zip(*samples))
ax.set_aspect('equal')
# create title from n_samples, center, and radius, usign fstring
ax.set_title(f'{n_samples} Samples on a axes alligned Parallelogram with bottom left corner {min_x}, {min_y} and top right corner {min_y}, {max_y}')
# draw also the parallelogram in red
ax.plot([min_x, max_x], [min_y, min_y], color='r')
ax.plot([min_x, max_x], [max_y, max_y], color='r')
ax.plot([min_x, min_x], [min_y, max_y], color='r')
ax.plot([max_x, max_x], [min_y, max_y], color='r')
ax.set_xlabel('X')
ax.set_ylabel('Y')
plt.show()



import numpy as np
from sklearn.metrics import silhouette_score
import matplotlib.pyplot as plt
from sklearn.cluster import AgglomerativeClustering

# Generate random data points
np.random.seed(0)
X = np.random.randn(1000, 3000)

# Perform hierarchical/agglomerative clustering and find the best clustering using silhouette score
best_score = -1
best_n_clusters = -1
best_labels = None
for n_clusters in range(2, 10):
    clusterer = AgglomerativeClustering(n_clusters=n_clusters)
    cluster_labels = clusterer.fit_predict(X)
    silhouette_avg = silhouette_score(X, cluster_labels)
    print("For n_clusters =", n_clusters, "The average silhouette_score is :", silhouette_avg)
    if silhouette_avg > best_score:
        best_score = silhouette_avg
        best_n_clusters = n_clusters
        best_labels = cluster_labels

# project the data points to a 2D space using PCA and color the points according to cluster labels
from sklearn.decomposition import PCA
pca = PCA(n_components=2).fit(X)
pca_2d = pca.transform(X)
plt.scatter(pca_2d[:, 0], pca_2d[:, 1], c=best_labels)
plt.show()





import mrdja.sampling as sampling
import matplotlib.pyplot as plt
import numpy as np
import mrdja.geometry as geometry

def get_parallelogram_vertices(center, normal1, normal2, length1, length2):
    vertex1 = center + normal1 * length1 / 2 + normal2 * length2 / 2
    vertex2 = center - normal1 * length1 / 2 + normal2 * length2 / 2
    vertex3 = center - normal1 * length1 / 2 - normal2 * length2 / 2
    vertex4 = center + normal1 * length1 / 2 - normal2 * length2 / 2
    return [vertex1, vertex2, vertex3, vertex4]

n_samples = 100
normal1 = (1, 1)
normal2 = (-1, 1)
center = (0, 0)
length1 = 5
length2 = 5
normal1 = np.array(normal1)
normal2 = np.array(normal2)
center = np.array(center)
samples = sampling.sampling_parallelogram_2d(n_samples, normal1, normal2, center, length1, length2)
fig, ax = plt.subplots()
ax.scatter(*zip(*samples))
ax.set_aspect('equal')
# create title from n_samples, center, and radius, usign fstring
ax.set_title(f'{n_samples} Samples on a Parallelogram with normal vectors {normal1} and {normal2}, center {center}, length1 {length1}, and length2 {length2}')

# Draw the parallelogram
vertices = geometry.get_parallelogram_2d_vertices(center, normal1, normal2, length1, length2)
vertices.append(vertices[0])  # Connect the last vertex with the first to close the shape
vertices = np.array(vertices)
ax.plot(vertices[:, 0], vertices[:, 1], color='r')

ax.set_xlabel('X')
ax.set_ylabel('Y')
plt.show()



import mrdja.ransac.coreransac as coreransac
import mrdja.ransac.coreransaccuda as coreransaccuda
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np
import random

'''
Line-guided plane detection in pointclouds.
Tomar, por ejemplo, 200 rectas y luego los 40 mejores por parejas. Eso son 780 parejas. Tomar el 5% mejor según el error de ajuste del plano.
Eso son 39 parejas. En total 239 veces que hay que mirar los inliers mirando todos los puntos.
Calcular el mejor plano, tomando RANSAC de open3d con 100.000 iteraciones.
Hacer varios experimentos, tanto de RANSAC con 239 o más iteraciones, como de open3d RANSAC con otros números de iteraciones.
'''

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

def get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed):
    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    '''
    audit_before_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_before_sanitizing"] = audit_before_sanitizing
    pcd = pointcloud.pointcloud_sanitize(pcd)
    audit_after_sanitizing = pointcloud.pointcloud_audit(pcd)
    dict_full_results["audit_after_sanitizing"] = audit_after_sanitizing
    '''
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    dict_full_results["ransac_iterations_results"] = []

    np.random.seed(seed)
    max_number_inliers = 0
    best_iteration_results = None
    for i in range(ransac_iterations):
        print("Iteration", i)
        dict_iteration_results = coreransac.get_ransac_line_iteration_results(np_points, threshold, number_pcd_points)
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

ransac_iterations = 200
threshold = 0.02
seed = 42

# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/office_1/office_1.ply"
filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"
RANSAC_data_from_file = get_RANSAC_data_from_file(filename, ransac_iterations, threshold, seed)

print(RANSAC_data_from_file)

# create a function that takes the RANSAC_data_from_file and the iteration number and returns the inliers
# create a function that takes the RANSAC_data_from_file and the iteration number and returns the number of inliers

pcd = o3d.io.read_point_cloud(RANSAC_data_from_file["filename"])
np_points = np.asarray(pcd.points)

# create a function that returns all the current_line along with their number_inliers

def get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file):
    pair_lines_number_inliers = []
    for dict_iteration_results in RANSAC_data_from_file["ransac_iterations_results"]:
        pair_lines_number_inliers.append((dict_iteration_results["current_line"], dict_iteration_results["number_inliers"]))
    return pair_lines_number_inliers

# order the pairs by number_inliers

def get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data_from_file):
    pair_lines_number_inliers = get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file)
    pair_lines_number_inliers_ordered = sorted(pair_lines_number_inliers, key=lambda pair_line_number_inliers: pair_line_number_inliers[1], reverse=True)
    return pair_lines_number_inliers_ordered

ordered_pairs = get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data_from_file)

list_sse_plane = []
current_iteration = 0
for i in range(40):
    for j in range(i+1, 40):
        # compute the current number of iterations
        current_iteration += 1
        print("Iteration", current_iteration)
        line_1 = ordered_pairs[i][0]
        line_2 = ordered_pairs[j][0]
        points = np.array([line_1[0], line_1[1], line_2[0], line_2[1]])
        a, b, c, d, sse = geometry.fit_plane_svd(points)
        new_plane = np.array([a, b, c, d])
        threshold_pcd = 0.02
        list_sse_plane.append((sse, new_plane))

# get the plane values for the 5% percentile of sse (the best planes)
list_sse = [pair_sse_plane[0] for pair_sse_plane in list_sse_plane]
list_plane = [pair_sse_plane[1] for pair_sse_plane in list_sse_plane]
percentile_05 = np.percentile(list_sse, 5)
list_sse_05 = [list_sse[i] for i in range(len(list_sse)) if list_sse[i] <= percentile_05]
list_plane_05 = [list_plane[i] for i in range(len(list_plane)) if list_sse[i] <= percentile_05]
print (list_sse_05)
print (list_plane_05)

# get the best plane from the 95% percentile of sse and the number of inliers

# Extract x, y, z coordinates from np_points
points_x = np_points[:, 0]
points_y = np_points[:, 1]
points_z = np_points[:, 2]

how_many_maximum = 0
best_plane = None
print("Starting the final loop")
for i in range(len(list_sse_05)):
    print("Iteration", i)
    new_plane = list_plane_05[i]
    threshold_pcd = 0.02
    how_many, inliers = coreransaccuda.get_how_many_below_threshold_between_plane_and_points_and_their_indices_cuda(np_points, 
                                                                                                points_x,
                                                                                                points_y,
                                                                                                points_z,
                                                                                                new_plane, threshold_pcd)

    if how_many > how_many_maximum:
        how_many_maximum = how_many
        inliers_maximum = inliers
        best_plane = new_plane
        sse_best_plane = list_sse_05[i]
print (how_many_maximum)
print (best_plane)
print (sse_best_plane)
inliers = np.asarray(inliers_maximum)
inlier_cloud = pcd.select_by_index(inliers)
inlier_cloud.paint_uniform_color([1, 0, 0])
o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers  " + str(how_many) + " inliers")

# set random seed
np.random.seed(42)
plane_model, inliers = pcd.segment_plane(distance_threshold=0.02,
                                         ransac_n=3,
                                         num_iterations=478)
[a, b, c, d] = plane_model
print("Plane equation: {:.2f}x + {:.2f}y + {:.2f}z + {:.2f} = 0".format(a, b, c, d))
inlier_cloud = pcd.select_by_index(inliers)
inlier_cloud.paint_uniform_color([1, 0, 0])
o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers  " + str(len(inliers)) + " inliers")
print("Number of inliers: " + str(len(inliers)))
print("Number of inliers maximum: " + str(how_many_maximum))






import os
import glob
import pickle

# Load the max_values dictionary
with open('/home/scpmaotj/Stanford3dDataset_v1.2/max_values.pkl', 'rb') as file:
    max_values = pickle.load(file)

def add_maximum_to_files(directory):
    for filepath in glob.glob(directory + '**/*.pkl', recursive=True):
        print(filepath)
        with open(filepath, 'rb') as file:
            data = pickle.load(file)
        relative_path = os.path.relpath(filepath, directory)
        key = relative_path.replace(os.path.sep, '_').replace('.pkl', '_maximum')
        #remove _means from key
        key = key.replace('_means', '')
        print("Key: ", key)
        if key in max_values:
            print(key)
            data['maximum'] = max_values[key]
            with open(filepath, 'wb') as file:
                pickle.dump(data, file)

directory = "/home/scpmaotj/pkl_files/me/" # specify your directory
add_maximum_to_files(directory)


import mrdja.ransac.coreransaccuda as coreransaccuda
import mrdja.ransac.coreransac as coreransac
import mrdja.pointcloud as pointcloud
import mrdja.geometry as geometry
import mrdja.drawing as drawing
import open3d as o3d
import numpy as np
import random
import pickle as pkl
from numba import cuda
import time

'''
Line-guided plane detection in pointclouds.
Tomar, por ejemplo, 200 rectas y luego los 40 mejores por parejas. Eso son 780 parejas. Tomar el 5% mejor según el error de ajuste del plano.
Eso son 39 parejas. En total 239 veces que hay que mirar los inliers mirando todos los puntos.
Calcular el mejor plano, tomando RANSAC de open3d con 100.000 iteraciones.
Hacer varios experimentos, tanto de RANSAC con 239 o más iteraciones, como de open3d RANSAC con otros números de iteraciones.
'''

# The files in the Stanford database are like this:
# {'number_pcd_points': 1136617, 'has_normals': False, 
# 'has_colors': True, 'is_empty': False, 'max_x': -15.207, 
# 'min_x': -20.542, 'max_y': 41.283, 'min_y': 36.802, 
# 'max_z': 3.206, 'min_z': 0.02, 'all_points_finite': True, 
# 'all_points_unique': False}
# the measures are in meters
# dict_results = pointcloud_audit(pcd)
# print(dict_results)

def get_RANSAC_data_from_file(filename, ransac_iterations, threshold):
    dict_full_results = {}
    dict_full_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    number_pcd_points = len(pcd.points)
    dict_full_results["number_pcd_points"] = number_pcd_points
    np_points = np.asarray(pcd.points)

    # Extract x, y, z coordinates from np_points
    points_x = np_points[:, 0]
    points_y = np_points[:, 1]
    points_z = np_points[:, 2]

    # Convert points_x, points_y, and points_z to contiguous arrays
    d_points_x = np.ascontiguousarray(points_x)
    d_points_y = np.ascontiguousarray(points_y)
    d_points_z = np.ascontiguousarray(points_z)

    d_points_x = cuda.to_device(d_points_x)
    d_points_y = cuda.to_device(d_points_y)
    d_points_z = cuda.to_device(d_points_z)

    dict_full_results["ransac_iterations_results"] = []

    # np.random.seed(seed)
    max_number_inliers = 0
    best_iteration_results = None
    random_numbers_pairs = coreransac.get_np_array_of_two_random_points_from_np_array_of_points(np_points, repetitions=ransac_iterations, num_points=number_pcd_points)

    for i in range(ransac_iterations):
        print("Iteration of line fitting", i, "out of", ransac_iterations)
        random_points = random_numbers_pairs[i]
        dict_iteration_results = coreransaccuda.get_ransac_line_iteration_results_cuda(np_points, d_points_x, d_points_y, d_points_z, number_pcd_points, threshold, random_points=random_points)
        if dict_iteration_results["number_inliers"] > max_number_inliers:
            max_number_inliers = dict_iteration_results["number_inliers"]
            best_iteration_results = dict_iteration_results
        dict_full_results["ransac_iterations_results"].append(dict_iteration_results)
    dict_full_results["ransac_best_iteration_results"] = best_iteration_results
    return dict_full_results

def compute_parameters_RANSAC_line (line_iterations):
    number_chosen_lines = int(line_iterations / 5)
    number_lines_pairs = int(number_chosen_lines * (number_chosen_lines - 1) / 2)
    number_chosen_planes = int(number_lines_pairs * 0.05)
    total_iterations = line_iterations + number_chosen_planes
    return number_chosen_lines, number_lines_pairs, number_chosen_planes, total_iterations

# create a function that returns all the current_line along with their number_inliers

def get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file):
    pair_lines_number_inliers = []
    for dict_iteration_results in RANSAC_data_from_file["ransac_iterations_results"]:
        pair_lines_number_inliers.append((dict_iteration_results["current_line"], dict_iteration_results["number_inliers"]))
    return pair_lines_number_inliers

# order the pairs by number_inliers

def get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data_from_file):
    pair_lines_number_inliers = get_lines_and_number_inliers_from_RANSAC_data_from_file(RANSAC_data_from_file)
    pair_lines_number_inliers_ordered = sorted(pair_lines_number_inliers, key=lambda pair_line_number_inliers: pair_line_number_inliers[1], reverse=True)
    return pair_lines_number_inliers_ordered


# get a list with all the ply files under database_path
import glob
import os

database_path = "/home/scpmaotj/Stanford3dDataset_v1.2/"
# database_path = "/home/bee/S3DIS/Stanford3dDataset_v1.2/"
ply_files = glob.glob(database_path + "/**/*.ply", recursive=True)

print(ply_files)


threshold = 0.02
# seed = 42
# filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"

# filename_test = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/WC_1/WC_1.ply"

# ply_files = [filename_test]

t1 = time.time()

for filename in ply_files:

    filename_pkl_check = filename.replace(".ply", ".pkl")
    # check if filename_pkl_check exists
    if os.path.isfile(filename_pkl_check):
        continue

    print("filename", filename)

    dict_all_results = {}
    dict_all_results["filename"] = filename

    pcd = o3d.io.read_point_cloud(filename)
    np_points = np.asarray(pcd.points)

    # Extract x, y, z coordinates from np_points
    points_x = np_points[:, 0]
    points_y = np_points[:, 1]
    points_z = np_points[:, 2]

    # Convert points_x, points_y, and points_z to contiguous arrays
    d_points_x = np.ascontiguousarray(points_x)
    d_points_y = np.ascontiguousarray(points_y)
    d_points_z = np.ascontiguousarray(points_z)

    d_points_x = cuda.to_device(d_points_x)
    d_points_y = cuda.to_device(d_points_y)
    d_points_z = cuda.to_device(d_points_z)

    for num_iterations in [100, 200, 300, 400, 500, 600]:
    # for num_iterations in [10, 11, 12]:
        number_chosen_lines, number_lines_pairs, number_chosen_planes, total_iterations = compute_parameters_RANSAC_line(num_iterations)

        # filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/conferenceRoom_1/conferenceRoom_1.ply"
        # filename = "/home/scpmaotj/Stanford3dDataset_v1.2/Area_1/office_1/office_1.ply"

        dict_standard_RANSAC_results_list = list()
        dict_line_RANSAC_results_list = list()

        for n_loop in range(10):

            RANSAC_data_from_file = get_RANSAC_data_from_file(filename, num_iterations, threshold)

            print(RANSAC_data_from_file)

            # create a function that takes the RANSAC_data_from_file and the iteration number and returns the inliers
            # create a function that takes the RANSAC_data_from_file and the iteration number and returns the number of inliers

            ordered_pairs = get_lines_and_number_inliers_ordered_by_number_inliers(RANSAC_data_from_file)

            list_sse_plane = []
            current_iteration = 0
            for i in range(number_chosen_lines):
                for j in range(i+1, number_chosen_lines):
                    # compute the current number of iterations
                    current_iteration += 1
                    print("Iteration pairs of lines", current_iteration, "out of", number_lines_pairs, "Loop", n_loop, "out of", 10)
                    line_1 = ordered_pairs[i][0]
                    line_2 = ordered_pairs[j][0]
                    points = np.array([line_1[0], line_1[1], line_2[0], line_2[1]])
                    a, b, c, d, sse = geometry.fit_plane_svd(points)
                    new_plane = np.array([a, b, c, d])
                    list_sse_plane.append((sse, new_plane))

            # get the plane values for the 5% percentile of sse (the best planes)
            list_sse = [pair_sse_plane[0] for pair_sse_plane in list_sse_plane]
            list_plane = [pair_sse_plane[1] for pair_sse_plane in list_sse_plane]
            percentile_05 = np.percentile(list_sse, 5)
            list_sse_05 = [list_sse[i] for i in range(len(list_sse)) if list_sse[i] <= percentile_05]
            list_plane_05 = [list_plane[i] for i in range(len(list_plane)) if list_sse[i] <= percentile_05]
            print (list_sse_05)
            print (list_plane_05)

            # get the best plane from the 95% percentile of sse and the number of inliers

            how_many_maximum = 0
            best_plane = None
            print("Starting the final loop")
            for i in range(len(list_sse_05)):
                print("Iteration best fitting planes", i, "out of", len(list_sse_05), "Loop", n_loop, "out of", 10)
                new_plane = list_plane_05[i]
                how_many = coreransaccuda.get_how_many_below_threshold_between_plane_and_points_cuda(np_points, d_points_x, d_points_y, d_points_z, new_plane, threshold)
                if how_many > how_many_maximum:
                    how_many_maximum = how_many
                    best_plane = new_plane
                    sse_best_plane = list_sse_05[i]
            print (how_many_maximum)
            print (best_plane)
            print (sse_best_plane)
            # o3d.visualization.draw_geometries([pcd, inlier_cloud], window_name="Inliers  " + str(how_many) + " inliers")

            # set random seed
            # np.random.seed(42)
            plane_model, inliers = pcd.segment_plane(distance_threshold=threshold,
                                                    ransac_n=3,
                                                    num_iterations=total_iterations)

            dict_standard_RANSAC_results = {"n_points_pcd": len(np_points), "n_inliers_maximum": len(inliers), "best_plane": plane_model, "threshold": threshold}
            dict_line_RANSAC_results = {"n_points_pcd": len(np_points), "n_inliers_maximum": how_many_maximum, "best_plane": best_plane, "threshold": threshold}
            dict_line_RANSAC_results_list.append(dict_line_RANSAC_results)
            dict_standard_RANSAC_results_list.append(dict_standard_RANSAC_results)

        dict_all_results["standard_RANSAC_" + str(total_iterations)] = dict_standard_RANSAC_results_list
        dict_all_results["line_RANSAC_" + str(total_iterations)] = dict_line_RANSAC_results_list
        # get the mean of n_inliers_maximum of the elements of the list dict_line_RANSAC_results_list
        list_n_inliers_maximum = [int(dict_line_RANSAC_results["n_inliers_maximum"]) for dict_line_RANSAC_results in dict_line_RANSAC_results_list]
        mean_n_inliers_maximum = np.mean(list_n_inliers_maximum)
        dict_all_results["mean_n_inliers_maximum_line_RANSAC_" + str(total_iterations)] = mean_n_inliers_maximum
        # get the mean of n_inliers_maximum of the elements of the list dict_standard_RANSAC_results_list
        list_n_inliers_maximum = [int(dict_standard_RANSAC_results["n_inliers_maximum"]) for dict_standard_RANSAC_results in dict_standard_RANSAC_results_list]
        mean_n_inliers_maximum = np.mean(list_n_inliers_maximum)
        dict_all_results["mean_n_inliers_maximum_standard_RANSAC_" + str(total_iterations)] = mean_n_inliers_maximum

    plane_model, inliers = pcd.segment_plane(distance_threshold=threshold,
                                            ransac_n=3,
                                            num_iterations=100000)

    dict_baseline_results = {"n_points_pcd": len(np_points), "n_inliers_maximum": len(inliers), "best_plane": plane_model, "threshold": threshold}
    dict_all_results["standard_RANSAC_100000"] = dict_baseline_results

    # save the results as a pickle file in the same folder as the filename file; to do so, just change the extension of the file to pkl
    filename_pkl = filename.replace(".ply", ".pkl")
    with open(filename_pkl, 'wb') as f:
        pkl.dump(dict_all_results, f)

t2 = time.time()
print("Time elapsed:", t2-t1)
  


'''
NOT WORKING!!!! ABANDONED TILL I FIND A GOOD WAY OF SHOWING 3D INTERACTIVE PLOTS IN GRADIO
'''


import gradio as gr
import mrdja.sampling as sampling
import mrdja.geometry as geometry
import matplotlib.pyplot as plt
from gradio.components import Number, Image
import matplotlib
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
matplotlib.use('Agg')

def plot_parallelogram(n_samples, normal1_x, normal1_y, normal1_z, normal2_x, normal2_y, normal2_z,
                       normal3_x, normal3_y, normal3_z, center_x, center_y, center_z, length1, length2, length3):
    samples = sampling.sampling_parallelogram_3d(n_samples=n_samples, normal1=(normal1_x, normal1_y, normal1_z), 
                                                 normal2=(normal2_x, normal2_y, normal2_z), 
                                                 normal3=(normal3_x, normal3_y, normal3_z),
                                                 center=(center_x, center_y, center_z), 
                                                 length1=length1, length2=length2, length3=length3)
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(samples[:, 0], samples[:, 1], samples[:, 2])

    center = np.array([center_x, center_y, center_z])
    normal1 = np.array([normal1_x, normal1_y, normal1_z])
    normal2 = np.array([normal2_x, normal2_y, normal2_z])
    normal3 = np.array([normal3_x, normal3_y, normal3_z])
    vertices = geometry.get_parallelogram_3d_vertices(center, normal1, normal2, normal3, length1, length2, length3)
    xlim_min = np.min(vertices[:, 0])
    xlim_max = np.max(vertices[:, 0])
    ylim_min = np.min(vertices[:, 1])
    ylim_max = np.max(vertices[:, 1])
    zlim_min = np.min(vertices[:, 2])
    zlim_max = np.max(vertices[:, 2])
    graph_limits = geometry.get_limits_of_3d_graph_from_limits_of_object(xlim_min, xlim_max, ylim_min, ylim_max, zlim_min, zlim_max)
    ax.set_xlim(graph_limits[0], graph_limits[1])  # Set x-axis limits
    ax.set_ylim(graph_limits[2], graph_limits[3])  # Set y-axis limits
    ax.set_zlim(graph_limits[4], graph_limits[5])  # Set z-axis limits
    # create title from n_samples, center, and radius, using f-string
    ax.set_title(f'{n_samples} Samples on a Parallelogram with normal vectors {normal1} and {normal2}, center {center}, length1 {length1}, and length2 {length2}')
    # draw also the parallelogram in red
    edges = [(0, 1), (1, 2), (2, 3), (3, 0), (4, 5), (5, 6),
            (6, 7), (7, 4), (0, 4), (1, 5), (2, 6), (3, 7)]
    # Plot the edges
    for edge in edges:
        ax.plot([vertices[edge[0]][0], vertices[edge[1]][0]],
                [vertices[edge[0]][1], vertices[edge[1]][1]],
                [vertices[edge[0]][2], vertices[edge[1]][2]], color='red')

    # Draw the X, Y and Z axes as dotted lines
    ax.plot([0, 0], [0, 0], [-graph_limits[5], graph_limits[5]], 'k--')
    ax.plot([0, 0], [-graph_limits[3], graph_limits[3]], [0, 0], 'k--')
    ax.plot([-graph_limits[1], graph_limits[1]], [0, 0], [0, 0], 'k--')

    # Draw the normals at a quarter of their corresponding length
    quarter_length1 = length1 / 8
    quarter_length2 = length2 / 8
    quarter_length3 = length3 / 8
    arrow_length1 = quarter_length1 / 2
    arrow_length2 = quarter_length2 / 2
    arrow_length3 = quarter_length3 / 2
    ax.arrow(center_x, center_y, center_z, 
                normal1_x * quarter_length1, normal1_y * quarter_length1, normal1_z * quarter_length1, 
                head_width=arrow_length1, head_length=arrow_length1, fc='b', ec='b')
    ax.arrow(center_x, center_y, center_z,
                normal2_x * quarter_length2, normal2_y * quarter_length2, normal2_z * quarter_length2,
                head_width=arrow_length2, head_length=arrow_length1, fc='b', ec='b')
    ax.arrow(center_x, center_y, center_z,
                normal3_x * quarter_length3, normal3_y * quarter_length3, normal3_z * quarter_length3,  
                head_width=arrow_length3, head_length=arrow_length1, fc='b', ec='b')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    fig.canvas.draw()
    img = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close(fig)
    return img

# Define the input and output interfaces for Gradio
inputs = [
    Number(value=100, precision=0, label="Number of Samples"),
    Number(value=1, label="X value of the first normal vector"),
    Number(value=0, label="Y value of the first normal vector"),
    Number(value=0, label="Z value of the first normal vector"),
    Number(value=0, label="X value of the second normal vector"),
    Number(value=1, label="Y value of the second normal vector"),
    Number(value=0, label="Z value of the second normal vector"),
    Number(value=0, label="X value of the third normal vector"),
    Number(value=0, label="Y value of the third normal vector"),
    Number(value=1, label="Z value of the third normal vector"),
    Number(value=0, label="X value of the center of the parallelogram"),
    Number(value=0, label="Y value of the center of the parallelogram"),
    Number(value=1, label="Z value of the center of the parallelogram"),
    Number(value=1, label="Length of the first side of the parallelogram"),
    Number(value=1, label="Length of the second side of the parallelogram"),
    Number(value=1, label="Length of the third side of the parallelogram")
]
outputs = [
    Image(label="Parallelogram Plot", type="pil")
]

# Create the Gradio app
app = gr.Interface(
    fn=plot_parallelogram, 
    inputs=inputs, 
    outputs=outputs,
    title="Sampling 3D Parallelograms",
    description="Generate random samples on a 3D parallelogram",
    allow_flagging="never",
)

# Run the app
app.launch()



import numpy as np
import open3d as o3d
import mrdja.geometry as geom
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def draw_plane_as_lines_open3d(A, B, C, D, size=10, line_color=[1, 0, 0]):
    # Define the vertices of the plane
    vertices = np.array([
        [-size, -size, -(D + A * -size + B * -size) / C],
        [size, -size, -(D + A * size + B * -size) / C],
        [size, size, -(D + A * size + B * size) / C],
        [-size, size, -(D + A * -size + B * size) / C]
    ])

    # Define the lines
    lines = np.array([
        [0, 1],
        [1, 2],
        [2, 3],
        [3, 0]
    ])

    # Create the LineSet
    lineset = o3d.geometry.LineSet()
    lineset.points = o3d.utility.Vector3dVector(vertices)
    lineset.lines = o3d.utility.Vector2iVector(lines)

    # Set the color for each line
    colors = [line_color for i in range(len(lines))]
    lineset.colors = o3d.utility.Vector3dVector(colors)

    return lineset


def draw_face_of_cube(plane: np.ndarray, cube_min: np.ndarray, cube_max: np.ndarray, color: str = "red", alpha: float = 0.5, ax = None):
    '''
    Draws a plane in 3D space using matplotlib.

    :param plane: A 4x1 numpy array containing the coefficients of the plane corresponding to the face.
    :type plane: numpy.ndarray
    :param cube_min: A 3x1 numpy array containing the minimum x, y, and z values of the cube.
    :type cube_min: numpy.ndarray
    :param cube_max: A 3x1 numpy array containing the maximum x, y, and z values of the cube.
    :type cube_max: numpy.ndarray
    :param color: The color of the plane.
    :type color: str
    :param alpha: The transparency of the plane.
    :type alpha: float
    :param ax: The matplotlib axis to draw on.
    :type ax: matplotlib.axes.Axes3D
    :return: None
    :rtype: None
    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    # Define the plane's normal vector
    normal = plane[:3]
    # Create a grid of points on the plane for visualization
    x_vals = np.linspace(cube_min[0], cube_max[0], 50)
    y_vals = np.linspace(cube_min[1], cube_max[1], 50)
    z_vals = np.linspace(cube_min[2], cube_max[2], 50)
    x_grid, y_grid = np.meshgrid(x_vals, y_vals)
    if normal[2] != 0:
        z_grid = (-normal[0] * x_grid - normal[1] * y_grid - plane[3]) / normal[2]
    elif normal[1] != 0:
        x_grid, z_grid = np.meshgrid(x_vals, z_vals)
        y_grid = (-normal[0] * x_grid - normal[2] * z_grid - plane[3]) / normal[1]
    else:
        y_grid, z_grid = np.meshgrid(y_vals, z_vals)
        x_grid = (-normal[1] * y_grid - normal[2] * z_grid - plane[3]) / normal[0]
    # Plot the plane
    ax.plot_surface(x_grid, y_grid, z_grid, alpha=alpha, color=color)
    if plot_inside_function:
        # Set axis labels
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        # Show the plot
        plt.show()
    else:
        return ax

def draw_cube(cube_min: np.ndarray, cube_max: np.ndarray, color: str = "red", alpha: float = 0.5, ax = None):
    '''
    Draws a cube in 3D space using matplotlib.

    :param cube_min: A 3x1 numpy array containing the minimum x, y, and z values of the cube.
    :type cube_min: numpy.ndarray
    :param cube_max: A 3x1 numpy array containing the maximum x, y, and z values of the cube.
    :type cube_max: numpy.ndarray
    :param color: The color of the cube.
    :type color: str
    :param alpha: The transparency of the cube.
    :type alpha: float
    :return: None
    :rtype: None

    :Example:

    ::

        >>> import mrdja.drawing as drawing
        >>> import numpy as np
        >>> cube_min = np.array([-2, -2, -1])
        >>> cube_max = np.array([1, 2, 2])
        >>> drawing.draw_cube(cube_min, cube_max, color="red", alpha=0.5)

    |drawing_draw_cube_example|

    .. |drawing_draw_cube_example| image:: ../../_static/images/drawing_draw_cube_example.png

    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    # Plane equations for the cube faces (front, back, top, bottom, left, right).
    planes = [
        (1, 0, 0, -cube_min[0]),  # Front face
        (-1, 0, 0, cube_max[0]),  # Back face
        (0, 1, 0, -cube_min[1]),  # Top face
        (0, -1, 0, cube_max[1]),  # Bottom face
        (0, 0, 1, -cube_min[2]),  # Left face
        (0, 0, -1, cube_max[2])   # Right face
    ]
    # Draw the cube faces
    for plane in planes:
        # draw a plane limited to the cube's bounds
        draw_face_of_cube(plane, cube_min, cube_max, color=color, alpha=alpha, ax=ax)
    if plot_inside_function:
        # Set axis labels
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        # Show the plot
        plt.show()
    else:
        return ax

def draw_line_extension_to_plane(line: np.ndarray, plane: np.ndarray, ax = None):
    '''
    Draws a line in 3D space, and the intersection point of the line and the plane, but this does not draw the plane.

    :param line: A 2x3 numpy array containing the endpoints of the line.
    :type line: numpy.ndarray
    :param plane: A 4x1 numpy array containing the coefficients of the plane.
    :type plane: numpy.ndarray
    :return: None
    :rtype: None

    :Example:

    ::

        >>> import mrdja.geometry as geom
        >>> import numpy as np 
        >>> import mrdja.drawing as drawing
        >>> import matplotlib.pyplot as plt
        >>> line = np.array([[0, 0, 0], [1, 1, 1]])
        >>> plane = np.array([0, 0, 1, -3])
        >>> intersection_point = geom.get_intersection_point_of_line_with_plane(line, plane)
        >>> intersection_point
        array([3., 3., 3.])
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111, projection='3d')
        >>> drawing.draw_line_extension_to_plane(line, plane, ax)
        >>> 
        >>> drawing.draw_line_extension_to_plane(line, plane)

    |drawing_draw_line_extension_to_plane_example|

    .. |drawing_draw_line_extension_to_plane_example| image:: ../../_static/images/drawing_draw_line_extension_to_plane_example.png

    '''
    plot_inside_function = False
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plot_inside_function = True
    intersection_point = geom.get_intersection_point_of_line_with_plane(line, plane)
    # Calculate the distances from line[0] and line[1] to the intersection point
    distance_to_line0 = np.linalg.norm(intersection_point - line[0])
    distance_to_line1 = np.linalg.norm(intersection_point - line[1])

    # Determine which point is farther away
    farther_point_index = 0 if distance_to_line0 > distance_to_line1 else 1
    farther_point = line[farther_point_index]

    # Plot the remaining portion of the line from the farther point to the intersection point
    intersection_x = intersection_point[0]
    intersection_y = intersection_point[1]
    intersection_z = intersection_point[2]
    farther_x = farther_point[0]
    farther_y = farther_point[1]
    farther_z = farther_point[2]
    ax.plot([farther_x, intersection_x], [farther_y, intersection_y], [farther_z, intersection_z], linestyle='--', color='blue')

    # Plot the intersection point
    ax.scatter(intersection_point[0], intersection_point[1], intersection_point[2], c='green', label='Intersection Point', s=100)

    if plot_inside_function:
        # Set axis labels
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')
        # Show the plot
        plt.show()
    else:
        return ax




import mrdja.sampling as sampling
import matplotlib.pyplot as plt

n_samples = 100
center = (2,3)
radius = 5
samples = sampling.sampling_circle_2d(n_samples=n_samples, center=center, radius=radius)
fig, ax = plt.subplots()
ax.scatter(*zip(*samples))
ax.set_aspect('equal')
# create title from n_samples, center, and radius, usign fstring
ax.set_title(f'{n_samples} Samples on a Circle with Center {center} and Radius {radius}')
# draw also the circle in red
circle = plt.Circle(center, radius, color='r', fill=False)
ax.add_artist(circle)
ax.set_xlabel('X')
ax.set_ylabel('Y')
plt.show()